from typing import Any
from uuid import UUID

from django.http import HttpResponse
from django.urls import reverse

from core.models import (
    Invitation,
    PlatformUser,
)

from backoffice.tests.base import BackOfficeTestCase


class InvitationTestCase(BackOfficeTestCase):
    def send_invitation(
        self,
        *,
        user: PlatformUser,
        requester_id: str | UUID,
        requested_id: str | UUID,
        invitation_type: Invitation.InvitationType,
        url: str | None = None,
        exprected_count: int = 1,
        expected_code: int = 302,
    ) -> HttpResponse:
        self.client.force_login(user)
        response = self.client.post(
            url
            if url
            else reverse(
                'backoffice:invitation-send',
                kwargs={
                    'type': invitation_type.lower(),
                    'requester_id': requester_id,
                },
            ),
            data={
                'requester_id': requester_id,
                'requested_id': requested_id,
            },
        )
        self.assertEqual(Invitation.objects.count(), exprected_count)
        self.assertEqual(response.status_code, expected_code)
        return response

    def send_successful_invitation(
        self,
        *,
        user: PlatformUser,
        invitation_type: Invitation.InvitationType,
        expected_code: int = 302,
        **kwargs: Any,
    ) -> tuple[HttpResponse, Invitation]:
        response = self.send_invitation(
            user=user,
            invitation_type=invitation_type,
            expected_code=expected_code,
            **kwargs,
        )
        invitation = Invitation.objects.filter(state=Invitation.RequestsState.REQUESTED).first()
        assert invitation
        self.assertEqual(invitation.sender, user)
        self.assertEqual(invitation.type, invitation_type)
        if expected_code == 302:
            self.assertRedirects(response, reverse('backoffice:invitation-detail', kwargs={'pk': invitation.id}))
        return (response, invitation)

    def send_action(
        self,
        *,
        user: PlatformUser,
        invitation: Invitation,
        action: str,
        expected_code: int = 302,
        expected_state: Invitation.RequestsState | None = None,
        expected_count: int = 1,
        data: dict[str, Any] | None = None,
    ) -> HttpResponse:
        data = data or {}
        if action == 'update':
            url = reverse('backoffice:invitation-edit', kwargs={'pk': invitation.id})
        else:
            data['action'] = action
            url = reverse('backoffice:invitation-decision', kwargs={'pk': invitation.id})

        self.client.force_login(user)
        response = self.client.post(
            url,
            data=data,
        )
        self.assertEqual(response.status_code, expected_code)
        self.assertEqual(Invitation.objects.filter(state=expected_state).count(), expected_count)
        if expected_code == 302:
            self.assertRedirects(response, reverse('backoffice:invitation-detail', kwargs={'pk': invitation.id}))
        if expected_state:
            invitation.refresh_from_db()
            self.assertEqual(invitation.state, expected_state)
            if expected_state != Invitation.RequestsState.REQUESTED:
                self.assertEqual(invitation.decision_by, user)
            else:
                self.assertIsNone(invitation.decision_by)
        return response
