from django.urls import path, re_path
from django.views.generic import RedirectView

from backoffice.views import (
    assemblies,
    assemblyteam,
    auth,
    badges,
    events,
    invitations,
    misc,
    moderation,
    profile,
    projects,
    schedules,
    teams,
    vouchers,
    wiki,
)
from backoffice.views.conferences import (
    ConferencePublicationView,
    ConferenceRegistrationView,
    ConferenceSelectionView,
)
from backoffice.views.map import (
    FloorCreateView,
    FloorListView,
    FloorUpdateView,
    POICreateView,
    POIListView,
    POIUpdateView,
)

app_name = 'backoffice'
urlpatterns = [
    path('', assemblies.AssemblyListView.as_view(), name='index'),
    path('accounts/profile/', profile.ProfileView.as_view(), name='profile'),
    path('accounts/signup/', auth.RegistrationView.as_view(), name='signup'),
    path('accounts/signup/done', auth.SignupDoneView.as_view(), name='account_activation_sent'),
    path('accounts/change-password/', auth.PasswordChangeView.as_view(), name='password_change'),
    path('accounts/change-password/done', auth.PasswordChangeDoneView.as_view(), name='password_change_done'),
    path('accounts/reset-password/', auth.PasswordResetView.as_view(), name='password_reset'),
    path('accounts/reset-password/done', auth.PasswordResetDoneView.as_view(), name='password_reset_done'),
    path('accounts/reset-password/confirm/<uidb64>/<token>/', auth.PasswordResetConfirmView.as_view(), name='password_reset_confirm'),
    path('accounts/reset-password/complete', auth.PasswordResetCompleteView.as_view(), name='password_reset_complete'),
    re_path(
        r'^accounts/activate/(?P<uid_b64>[0-9A-Za-z_\-]+)/(?P<channel_id>\d+)/(?P<token>[0-9A-Za-z]{1,13}-[0-9A-Za-z]{1,40})/$',
        auth.RegistrationActivationView.as_view(),
        name='signup_activate',
    ),  # noqa: E501
    path('login', auth.LoginView.as_view(), name='login'),
    path('logout', auth.LogoutView.as_view(), name='logout'),
    path('auth_debug', auth.AuthDebugView.as_view()),
    path('conferences', ConferenceSelectionView.as_view(), name='conferences'),
    path('conference/<uuid:pk>', ConferencePublicationView.as_view(), name='conference-publication'),
    path('conference/<uuid:pk>/registration', ConferenceRegistrationView.as_view(), name='conference-registration'),
    path('wiki', wiki.WikiOverviewView.as_view(), name='wiki'),
    path('wiki/namespaces', wiki.NamespaceListView.as_view(), name='wiki-namespaces'),
    path('wiki/locks', wiki.LockListView.as_view(), name='wiki-locks'),
    path('wiki/pages', wiki.PagesView.as_view(), name='wiki-pages'),
    path('wiki/page/<uuid:pk>', wiki.PageView.as_view(), name='wiki-page-detail'),
    path('wiki/page/<uuid:pk>/delete', wiki.PageDeleteView.as_view(), name='wiki-page-delete'),
    path('wiki/page/<uuid:pk>/delete-revision', wiki.PageRevisionDeleteView.as_view(), name='wiki-page-revision-delete'),
    path('assemblies', assemblyteam.AssembliesView.as_view(), name='assemblies'),
    path('assemblies/floor.geojson', assemblies.AssemblyGeoJsonExportView.as_view(), name='assemblies-floor-geojson'),
    path('assemblies/list/<str:variant>', assemblyteam.AssembliesListsView.as_view(), name='assemblieslist'),
    path('assemblyteam/<uuid:pk>', assemblyteam.AssemblyView.as_view(), name='assemblyteam-detail'),
    path('assemblyteam/<uuid:pk>/state', assemblyteam.AssemblyEditStateView.as_view(), name='assemblyteam-editstate'),
    path('assemblyteam/<uuid:pk>/hierarchy', assemblyteam.AssemblyEditHierarchyView.as_view(), name='assemblyteam-edithierarchy'),
    path('assemblyteam/<uuid:pk>/position', assemblyteam.AssemblyEditPlacementView.as_view(), name='assemblyteam-editposition-assembly'),
    path('assemblyteam/<uuid:pk>/message', assemblyteam.AssemblyMessageView.as_view(), name='assemblyteam-message'),
    path('assemblyteam/room/<uuid:pk>/position', assemblyteam.EditRoomPositionView.as_view(), name='assemblyteam-editposition-room'),
    path('assembly/create', assemblies.AssemblyCreateView.as_view(), name='assembly-create'),
    path('assembly/<uuid:pk>', assemblies.AssemblyDetailView.as_view(), name='assembly'),
    path('assembly/<uuid:pk>/edit', assemblies.AssemblyUpdateView.as_view(), name='assembly-edit'),
    path('assembly/<uuid:pk>/editchildren', assemblies.AssemblyChildrenUpdateView.as_view(), name='assembly-editchildren'),
    path('assembly/<uuid:pk>/editlinks', assemblies.AssemblyLinksUpdateView.as_view(), name='assembly-editlinks'),
    path('assembly/<uuid:pk>/members', assemblies.MemberListView.as_view(), name='assembly-members'),
    path('assembly/<uuid:pk>/members/add', assemblies.MemberCreateView.as_view(), name='assembly-members-add'),
    path('assembly/<uuid:pk>/members/edit/<str:uname>', assemblies.MembersUpdateView.as_view(), name='assembly-members-edit'),
    path('assembly/<uuid:pk>/vouchers', assemblies.VouchersView.as_view(), name='assembly-vouchers'),
    path('assembly/<uuid:pk>/parent/leave', assemblies.AssemblyParentLeaveView.as_view(), name='assembly-parent-leave'),
    path('assembly/<uuid:assembly>/auth', assemblies.AuthView.as_view(), name='assembly-auth'),
    path('assembly/<uuid:assembly>/auth/app/<int:pk>', assemblies.AuthAppView.as_view(), name='assembly-auth-app'),
    path('assembly/<uuid:assembly>/auth/new_token', assemblies.AuthGetTokenView.as_view(), name='assembly-auth-gettoken'),
    path('assembly/<uuid:assembly>/badges', badges.BadgeListView.as_view(), name='assembly-badges'),
    path('assembly/<uuid:assembly>/badge/new', badges.BadgeCreateView.as_view(), name='assembly-create-badge'),
    path('assembly/<uuid:assembly>/badge/<uuid:pk>', badges.BadgeUpdateView.as_view(), name='assembly-badge'),
    path('assembly/<uuid:assembly>/badge/<uuid:pk>/renew_token', badges.BadgeRenewIssuingTokenView.as_view(), name='assembly-badge-renew'),
    path('assembly/<uuid:assembly>/badge/<uuid:pk>/remove', badges.BadgeDeleteView.as_view(), name='assembly-badge-remove'),
    path('assembly/<uuid:assembly>/badge/<uuid:pk>/award', badges.BadgeAwardView.as_view(), name='assembly-badge-award'),
    path('assembly/<uuid:assembly>/badge/<uuid:badge>/redeem_token/new', badges.BadgeTokenCreateView.as_view(), name='assembly-badge-create-redeem-token'),
    path('assembly/<uuid:assembly>/badge/<uuid:badge>/redeem_token/<uuid:pk>', badges.BadgeTokenUpdateView.as_view(), name='assembly-badge-redeem-token'),
    path(
        'assembly/<uuid:assembly>/badge/<uuid:badge>/redeem_token/<uuid:pk>/switch',
        badges.BadgeTokenToggleActiveView.as_view(),
        name='assembly-badge-toggle-active-redeem-token',
    ),
    path('assembly/<uuid:assembly>/new_event', events.AssemblyEventCreateView.as_view(), name='assembly-create-event'),
    path('assembly/<uuid:assembly>/events', events.AssemblyEventListView.as_view(), name='assembly-events'),
    path('assembly/<uuid:assembly>/new_room', assemblies.AssemblyRoomCreateView.as_view(), name='assembly-create-room'),
    path('assembly/<uuid:assembly>/new_project', projects.AssemblyProjectCreateView.as_view(), name='assembly-create-project'),
    path('assembly/<uuid:assembly>/projects', projects.AssemblyProjectListView.as_view(), name='assembly-projects'),
    path('assembly/<uuid:assembly>/p/<uuid:pk>/', projects.AssemblyProjectUpdateView.as_view(), name='assembly-project'),
    path('assembly/<uuid:assembly>/p/<uuid:pk>/remove', projects.ProjectDeleteView.as_view(), name='assembly-remove-project'),
    path('assembly/<uuid:assembly>/e/<uuid:pk>/', events.AssemblyEventUpdateView.as_view(), name='assembly-event'),
    path('assembly/<uuid:assembly>/e/<uuid:pk>/remove', events.AssemblyEventDeleteView.as_view(), name='assembly-event-remove'),
    path('assembly/<uuid:assembly>/r/<uuid:pk>/', assemblies.AssemblyRoomUpdateView.as_view(), name='assembly-room'),
    path('assembly/<uuid:assembly>/r/<uuid:room>/new_link', assemblies.RoomLinkCreateView.as_view(), name='roomlink-create'),
    path('assembly/<uuid:assembly>/r/<uuid:room>/remove_link', assemblies.RoomLinkDeleteView.as_view(), name='roomlink-remove'),
    path('assembly/<uuid:assembly>/r/<uuid:room>/remove', assemblies.AssemblyRoomDeleteView.as_view(), name='assembly-remove-room'),
    path('invitations', invitations.InvitationListView.as_view(), name='invitations'),
    path('invitation/<uuid:pk>', invitations.InvitationDetailView.as_view(), name='invitation-detail'),
    path('invitation/<uuid:pk>/edit', invitations.InvitationUpdateView.as_view(), name='invitation-edit'),
    path('invitation/<uuid:pk>/decision', invitations.InvitationDecisionView.as_view(), name='invitation-decision'),
    path('invite/<str:type>/<uuid:requester_id>', invitations.InvitationCreateView.as_view(), name='invitation-send'),
    path('invite/<str:type>/<uuid:requester_id>/<uuid:requested_id>', invitations.InvitationCreateView.as_view(), name='invitation-send'),
    path('map/floors', FloorListView.as_view(), name='map-floor-list'),
    path('map/floor/new', FloorCreateView.as_view(), name='map-floor-create'),
    path('map/floor/<uuid:pk>', FloorUpdateView.as_view(), name='map-floor-edit'),
    path('map/pois', POIListView.as_view(), name='map-poi-list'),
    path('map/poi/new', POICreateView.as_view(), name='map-poi-create'),
    path('map/poi/<uuid:pk>', POIUpdateView.as_view(), name='map-poi-edit'),
    path('moderation', moderation.IndexView.as_view(), name='moderation'),
    path('moderation/assemblies', moderation.ModerationAssemblyListView.as_view(), name='moderation-assembly-list'),
    path('moderation/assembly/<uuid:pk>', moderation.ModerationAssemblyDetailView.as_view(), name='moderation-assembly-detail'),
    path('moderation/events', moderation.ModerationEventListView.as_view(), name='moderation-event-list'),
    path('moderation/event/<uuid:pk>', moderation.ModerationEventDetailView.as_view(), name='moderation-event-detail'),
    path('moderation/badges', moderation.ModerationBadgeListView.as_view(), name='moderation-badge-list'),
    path('moderation/badge/<uuid:pk>', moderation.ModerationBadgeDetailView.as_view(), name='moderation-badge-detail'),
    path('moderation/board', moderation.ModerationBoardListView.as_view(), name='moderation-board-list'),
    path('moderation/board/<int:pk>', moderation.ModerationBoardDetailView.as_view(), name='moderation-board-detail'),
    path('moderation/project', moderation.ModerationProjectListView.as_view(), name='moderation-project-list'),
    path('moderation/project/<uuid:pk>', moderation.ModerationProjectDetailView.as_view(), name='moderation-project-detail'),
    path('moderation/users', moderation.ModerationUserListView.as_view(), name='moderation-user-list'),
    path('moderation/user/<int:pk>', moderation.ModerationUserDetailView.as_view(), name='moderation-user-detail'),
    path('moderation/user/<int:pk>/rename', moderation.ModerationUserRenameView.as_view(), name='moderation-user-rename'),
    path('moderation/wiki', moderation.ModerationWikiListView.as_view(), name='moderation-wiki-list'),
    path('moderation/wiki/<uuid:pk>', moderation.ModerationWikiDetailView.as_view(), name='moderation-wiki-detail'),
    path('schedule/', schedules.SchedulesIndexView.as_view(), name='schedules'),
    path('schedule/sources', schedules.ScheduleSourcesListView.as_view(), name='schedulesource-list'),
    path('schedule/source/add', schedules.ScheduleSourcesCreateView.as_view(), name='schedulesource-add'),
    path('schedule/source/<uuid:pk>', schedules.ScheduleSourcesDetailView.as_view(), name='schedulesource-detail'),
    path('schedule/source/<uuid:pk>/delete', schedules.ScheduleSourcesDeleteView.as_view(), name='schedulesource-remove'),
    path('schedule/source/<uuid:pk>/import', schedules.ScheduleSourcesDoImportView.as_view(), name='schedulesource-import'),
    path('schedule/source/<uuid:pk>/update', schedules.ScheduleSourcesUpdateView.as_view(), name='schedulesource-edit'),
    path('schedule/import/<int:pk>', schedules.ScheduleSourceImportDetailView.as_view(), name='schedulesourceimport-detail'),
    path('schedule/import/<int:pk>.json', schedules.ScheduleSourceImportDataView.as_view(), name='schedulesourceimport-detail-data'),
    path('sos/', RedirectView.as_view(pattern_name='backoffice:sos', permanent=True)),
    path('self-organized/', events.SelfOrganizedContentListView.as_view(), name='sos'),
    path('self-organized/projects/create', projects.ProjectCreateView.as_view(), name='so-create-project'),
    path('self-organized/projects/<uuid:pk>/', projects.SelfOrganizedProjectUpdateView.as_view(), name='so-project'),
    path('self-organized/projects/<uuid:pk>/delete', projects.ProjectDeleteView.as_view(), name='so-remove-project'),
    path('self-organized/sessions/new', events.SoSCreateView.as_view(), name='sos-create'),
    path('self-organized/sessions/<uuid:pk>/', events.SoSUpdateView.as_view(), name='sos-edit'),
    path('self-organized/sessions/<uuid:pk>/delete', events.SoSDeleteView.as_view(), name='sos-delete'),
    path('sos/new', RedirectView.as_view(pattern_name='backoffice:sos-create')),
    path('sos/<uuid:pk>/', RedirectView.as_view(pattern_name='backoffice:sos-edit')),
    path('teams', teams.TeamListView.as_view(), name='teams'),
    path('team/create', teams.TeamCreateView.as_view(), name='team-create'),
    path('team/<uuid:uuid>', teams.TeamDetailView.as_view(), name='team'),
    path('team/<uuid:uuid>/edit', teams.TeamUpdateView.as_view(), name='team-edit'),
    path('team/<uuid:uuid>/delete', teams.TeamDeleteView.as_view(), name='team-delete'),
    path('team/<uuid:team>/member/<uuid:pk>/delete', teams.TeamMemberDeleteView.as_view(), name='team-member-delete'),
    path('team/<uuid:team>/member/<uuid:pk>/update', teams.TeamMemberUpdateView.as_view(), name='team-member-update'),
    path('vouchers', vouchers.VouchersView.as_view(), name='vouchers'),
    path('_boom', misc.BoomView.as_view()),
]
