import logging

from django.contrib import messages
from django.urls import reverse
from django.utils.translation import gettext_lazy as _
from django.views.generic import ListView
from django.views.generic.edit import CreateView, DeleteView, UpdateView

from core.models.assemblies import Assembly
from core.models.conference import ConferenceExportCache
from core.models.events import Event

from backoffice.views.events.mixins import AssemblyEventFormMixin, EventPublicationMixin
from backoffice.views.mixins import AssemblyMixin

logger = logging.getLogger(__name__)


class AssemblyEventListView(AssemblyMixin, ListView):
    model = Assembly
    template_name = 'backoffice/event/list.html'
    context_object_name = 'events'

    assembly_url_param = 'assembly'

    def get_queryset(self, *args, **kwargs):
        return Event.objects.manageable_by_user(user=self.request.user, conference=self.conference).filter(assembly=self.assembly)

    def get_context_data(self, *args, **kwargs):
        return {
            'user_list': True,
            **super().get_context_data(*args, **kwargs),
        }


class AssemblyEventCreateView(AssemblyEventFormMixin, AssemblyMixin, CreateView):
    def get_form_kwargs(self, *args, **kwargs):
        return {'create': True, **super().get_form_kwargs(*args, **kwargs)}


class AssemblyEventUpdateView(EventPublicationMixin, AssemblyEventFormMixin, AssemblyMixin, UpdateView):
    def get_queryset(self, *args, **kwargs):
        return Event.objects.manageable_by_user(user=self.request.user, conference=self.conference).filter(assembly=self.assembly)


class AssemblyEventDeleteView(AssemblyMixin, DeleteView):
    assembly_url_param = 'assembly'
    assembly_management = True
    model = Event

    def get_object(self, *args, **kwargs):
        obj = super().get_object(*args, **kwargs)
        if obj.assembly != self.assembly:
            raise self.model.DoesNotExist
        return obj

    def form_valid(self, *args, **kwargs):
        object_id = self.object.id
        result = super().form_valid(*args, **kwargs)
        messages.success(self.request, _('Event__removed %(event_id)s') % {'event_id': object_id})
        logger.info('Event %(object_id)s removed by %(user)s', {'object_id': object_id, 'user': self.request.user})
        ConferenceExportCache.signal_schedule_modification(self.conference)
        return result

    def get_success_url(self, *args, **kwargs):
        return reverse('backoffice:assembly-events', kwargs={'assembly': self.assembly.id})
