import json
import uuid
from datetime import datetime
from pathlib import Path
from typing import TypeIs
from uuid import uuid4

from django.conf import settings
from django.contrib.contenttypes.fields import GenericRelation
from django.contrib.postgres.fields import DateTimeRangeField
from django.core.exceptions import ValidationError
from django.db import models
from django.db.models import ObjectDoesNotExist, Q, QuerySet
from django.utils import timezone
from django.utils.text import slugify
from django.utils.translation import gettext_lazy as _

from core.fields import ConferenceReference
from core.markdown import compile_translated_markdown_fields, store_relationships
from core.models.activitylog import ActivityLogMixin
from core.models.assemblies import Assembly
from core.models.base_managers import ConferenceManagerMixin
from core.models.conference import Conference, ConferenceMember
from core.models.shared import BackendMixin
from core.models.tags import TagItem
from core.utils import resolve_internal_url, resolve_link, str2bool
from core.validators import FileSizeValidator, ImageDimensionValidator


class RoomType(models.TextChoices):
    LECTURE_HALL = 'lecturehall', _('Room__type__lecture_hall')
    STAGE = 'stage', _('Room__type__stage')
    WORKSHOP = 'workshop', _('Room__type__workshop')
    OUTSIDE = 'outside', _('Room__type__outside')
    ONLINE = 'online', _('Room__type__online')
    PROJECT = 'project', _('Room__type__project')
    BIGBLUEBUTTON = 'bbb', _('Room__type__bbb')
    HANGAR = 'hangar', _('Room__type__hangar')
    OTHER = 'other', _('Room__type__other')


class RoomManager(ConferenceManagerMixin['Room']):
    staff_permissions = ['core.assembly_team']
    conference_filter = 'assembly__conference'
    assembly_filter = 'assembly'

    def apply_public_filter(self, queryset: 'QuerySet[Room]', member: ConferenceMember | None = None) -> 'QuerySet[Room]':
        return queryset.filter(blocked=False).filter(Q(assembly__state__in=Assembly.PUBLIC_STATES))

    def assignable_in_timeframe(self, conference: Conference, start: datetime, end: datetime, assembly: Assembly | None = None, is_sos: bool = False):
        """
        Fetch all rooms which are bookable for events in the given assembly (if any), in the given timeframe.
        Optionally, include rooms which are free for SOS usage.
        """
        additional_pk = None
        shares_qs = Q()

        # simple case: nothing to select
        if not assembly and not is_sos:
            return Room.objects.none()

        # we are looking for rooms assignable to events belonging to the given assembly
        if assembly:
            # sanity check against drunk devs
            if assembly.conference_id != conference.id:
                raise ValidationError('The given assembly must belong to the given conference.')

            # we want all rooms belonging to the assembly in question
            additional_pk = assembly.rooms.values_list('id', flat=True)

            # include all rooms shared with this assembly (or all assemblies)
            shares_qs |= Q(single_assembly=assembly)
            shares_qs |= Q(all_assemblies=True)

        # if we're searching for a SOS include all rooms which allow
        if is_sos:
            shares_qs |= Q(all_self_organized_sessions=True)

        # limit the shared rooms to either matching timeframe spec or none at all
        shares_qs &= Q(timeframe=None) | Q(timeframe__contains=(start, end))

        # select all rooms matching the sharing condition
        q = Q(shares__pk__in=RoomShare.objects.filter(enabled=True).filter(shares_qs))

        # combine the additional rooms, if given
        if additional_pk:
            q |= Q(pk__in=additional_pk)

        # construct the actual queryset
        result = conference.rooms.filter(room_type__in=Room.TYPES_ASSIGNABLE_TO_EVENTS).filter(q).distinct()
        return result


def get_banner_file_name(instance: 'Assembly', filename: str) -> str:
    return str(Path('rooms_projects') / str(instance.id) / 'banner' / filename)


def get_banner_help_test() -> str:
    return _('Room__banner_image__help %(limit)d, %(max_width)d, %(max_height)d') % {
        'limit': settings.ASSEMBLY_BANNER_FILE_SIZE_LIMIT,
        'max_width': settings.ASSEMBLY_BANNER_WIDTH_MAXIMUM,
        'max_height': settings.ASSEMBLY_BANNER_HEIGHT_MAXIMUM,
    }


class Room(BackendMixin, ActivityLogMixin, models.Model):
    class Meta:
        verbose_name = _('Room')
        verbose_name_plural = _('Rooms')
        constraints = [models.UniqueConstraint(fields=['assembly', 'name'], name='room_unique_name')]

    RoomType = RoomType

    class RecordingState(models.TextChoices):
        RECORD_BY_DEFAULT = 'record_by_default', _('Room__recording_state-record_by_default')
        NOT_RECORDED_BY_DEFAULT = 'not_recorded_by_default', _('Room__recording_state-not_recorded_by_default')
        RECORDING_FORBIDDEN = 'recording_forbidden', _('Room__recording_state-recording_forbidden')
        RECORDING_NOT_POSSIBLE = 'recording_not_possible', _('Room__recording_state-recording_not_possible')
        UNKNOWN = 'unknown', _('Room__recording_state-unknown')

    BACKEND_ROOMTYPES = [RoomType.BIGBLUEBUTTON, RoomType.HANGAR]
    TYPES_WITH_CAPACITY = [RoomType.LECTURE_HALL, RoomType.STAGE, RoomType.WORKSHOP, RoomType.BIGBLUEBUTTON]
    TYPES_WITH_LINKS = [RoomType.PROJECT, RoomType.LECTURE_HALL, RoomType.STAGE, RoomType.WORKSHOP, RoomType.OUTSIDE, RoomType.ONLINE, RoomType.OTHER]
    TYPES_ASSIGNABLE_TO_EVENTS = [
        RoomType.LECTURE_HALL,
        RoomType.STAGE,
        RoomType.WORKSHOP,
        RoomType.OUTSIDE,
        RoomType.ONLINE,
        RoomType.BIGBLUEBUTTON,
    ]

    objects = RoomManager()

    id = models.UUIDField(default=uuid4, primary_key=True, editable=False)
    conference = ConferenceReference(related_name='rooms')
    assembly = models.ForeignKey(Assembly, related_name='rooms', on_delete=models.CASCADE)

    blocked = models.BooleanField(default=False, help_text=_('Room__blocked__help'), verbose_name=_('Room__blocked'))
    """The room has been blocked for viewing."""

    name = models.CharField(max_length=200)
    """Name of the room as it can be found in the assembly."""

    slug = models.SlugField(unique=True)

    description = models.TextField(blank=True, null=True, help_text=_('Room__description__help'), verbose_name=_('Room__description'))
    """Description of the room/project."""
    description_html = models.TextField(blank=True, null=True)

    is_public_fahrplan = models.BooleanField(default=False, help_text=_('Room__is_public_fahrplan__help'), verbose_name=_('Room__is_public_fahrplan'))

    is_official = models.BooleanField(default=False, help_text=_('Room__is_official__help'), verbose_name=_('Room__is_official'))
    official_room_order = models.IntegerField(default=1000, help_text=_('Room__official_room_order__help'), verbose_name=_('Room__official_room_order'))

    room_type = models.CharField(max_length=20, choices=RoomType.choices, help_text=_('Room__type__help'), verbose_name=_('Room__type'))
    """Style of the room."""

    location_state = models.CharField(
        max_length=20,
        choices=Assembly.LocationState.choices,
        default=Assembly.LocationState.NONE,
        help_text=_('Room__location_state__help'),
        verbose_name=_('Room__location_state'),
    )
    location_floor = models.ForeignKey(
        'MapFloor',
        blank=True,
        null=True,
        related_name='rooms',
        on_delete=models.PROTECT,
        help_text=_('Room__location_floor__help'),
        verbose_name=_('Room__location_floor'),
    )
    location_data = models.JSONField(
        blank=True,
        null=True,
    )

    backend_link = models.URLField(blank=True, null=True, help_text=_('Room__backend_link__help'), verbose_name=_('Room__backend_link'))
    """Reference to integration/source system. For e.g. WorkAdventure this is the collection JSON URL."""

    backend_link_branch = models.CharField(
        max_length=50,
        blank=True,
        null=True,
        help_text=_('Room__backend_link_branch__help'),
        verbose_name=_('Room__backend_link_branch'),
    )
    """Additional reference to integration/source system, i.e. a git branch name."""

    backend_flags = models.JSONField(blank=True, null=True, help_text=_('Room__backend_flags__help'), verbose_name=_('Room__backend_flags'))
    """Arbitrary data which can be used to integration component."""

    backend_data = models.JSONField(blank=True, null=True, help_text=_('Room__backend_data__help'), verbose_name=_('Room__backend_data'))
    """Arbitrary data necessary for room setup in the backend - used by the actual integration component (not for users!!!)."""

    director_data = models.JSONField(blank=True, null=True, help_text=_('Room__director_data__help'), verbose_name=_('Room__director_data'))
    """Arbitrary data necessary for room operation in the backend - used by the actual integration component (not for users!!!)."""

    backend_status = models.CharField(
        max_length=40,
        blank=True,
        choices=BackendMixin.BackendStatus.choices,
        default=BackendMixin.BackendStatus.NONE,
        help_text=_('Room__backend_status__help'),
        verbose_name=_('Room__backend_status'),
    )
    """The backend's status information."""

    capacity = models.PositiveIntegerField(blank=True, null=True, help_text=_('Room__capacity__help'), verbose_name=_('Room__capacity'))
    """A room's capacity, i.e. maximum of participants."""

    reserve_capacity = models.PositiveIntegerField(
        default=20, help_text=_('Room__reserve_capacity__help'), verbose_name=_('Room__reserve_capacity'), editable=False
    )
    """Number of slots the deployment infrastructure should keep available for this room"""

    occupants = models.PositiveIntegerField(blank=True, null=True, help_text=_('Room__occupants__help'), verbose_name=_('Room__occupants'), editable=False)
    """The (estimated) number of occupants in the room as reported by the backend."""
    recording_state = models.CharField(
        max_length=40,
        choices=RecordingState.choices,
        default=RecordingState.UNKNOWN,
        help_text=_('Room__recording_state__help'),
        verbose_name=_('Room__recording_state'),
    )

    last_update = models.DateTimeField(auto_now=True)

    tags = GenericRelation(TagItem, object_id_field='target_id', content_type_field='target_type', related_query_name='rooms')

    banner_image_height = models.PositiveIntegerField(blank=True, null=True)
    banner_image_width = models.PositiveIntegerField(blank=True, null=True)
    banner_image = models.ImageField(
        blank=True,
        null=True,
        height_field='banner_image_height',
        width_field='banner_image_width',
        help_text=get_banner_help_test,
        upload_to=get_banner_file_name,
        verbose_name=_('Assembly__banner_image'),
        validators=[
            FileSizeValidator(settings.ASSEMBLY_BANNER_FILE_SIZE_LIMIT),
            ImageDimensionValidator(
                min_size=(
                    settings.ASSEMBLY_BANNER_WIDTH_MINIMUM,
                    settings.ASSEMBLY_BANNER_HEIGHT_MINIMUM,
                ),
                max_size=(
                    settings.ASSEMBLY_BANNER_WIDTH_MAXIMUM,
                    settings.ASSEMBLY_BANNER_HEIGHT_MAXIMUM,
                ),
            ),
        ],
    )

    def __str__(self):
        return self.name

    @classmethod
    def type_is(cls, obj: object) -> TypeIs['Room']:
        return isinstance(obj, cls)

    @property
    def conference_slug(self):
        return self.conference.slug

    @property
    def assembly_slug(self):
        return self.assembly.slug if self.assembly is not None else None

    @property
    def has_backend(self):
        return self.room_type in self.BACKEND_ROOMTYPES

    @property
    def backend_status_class(self):
        if self.backend_status in [self.BackendStatus.ACTIVE]:
            return 'good'

        if self.backend_status in [self.BackendStatus.FULL, self.BackendStatus.INACTIVE]:
            return 'warning'

        if self.backend_status in [self.BackendStatus.ERROR]:
            return 'bad'

        if self.backend_status in [self.BackendStatus.NEW, self.BackendStatus.SETUP]:
            return 'info'

        return 'unknown'

    @property
    def voc_stream_id(self):
        link = self.links.filter(link_type=RoomLink.LinkType.MEDIA_CCC_DE).first()
        if link:
            return link.link
        return None

    def get_absolute_url(self):
        from core.templatetags.hub_absolute import hub_absolute  # pylint: disable=import-outside-toplevel

        return hub_absolute('plainui:room', slug=self.slug, i18n=settings.ARCHIVE_MODE)

    def __create_slug(self, extension='', max_length: int = 50):
        """
        recursive function to generate a free room slug based on the room name
        """
        sluged = slugify(self.name + extension)
        if max_length and len(sluged) > max_length:
            sluged = sluged[:max_length]

        exists = False
        if self.id:
            exists = Room.objects.exclude(id=self.id).filter(slug=sluged).exists()
        else:
            exists = Room.objects.filter(slug=sluged).exists()

        if exists:
            extension = str(uuid4())[13:18]
            return self.__create_slug(extension)
        else:
            return sluged

    def generate_slug(self):
        self.slug = self.__create_slug()

    def get_location_floor_index(self) -> int | None:
        return self.location_floor.index if self.location_floor else None

    def get_location_point_xy(self) -> tuple[int, int] | None:
        if self.location_data is None:
            return None
        if p := self.location_data.get('point'):
            return p[0], p[1]
        return None

    def get_location_point_as_json(self) -> str:
        return json.dumps(self.get_location_point_xy() or '')

    def get_location_boundaries_xy(self) -> list[list[tuple[int, int]]] | None:
        if self.location_data is None:
            return None
        if b := self.location_data.get('boundaries'):
            return b
        return None

    def get_location_boundaries_as_json(self) -> str:
        return json.dumps(self.get_location_boundaries_xy() or '')

    def clean(self):
        if self.room_type == self.RoomType.PROJECT:
            raise ValidationError(_('Room__type_project_not_allowed'))

    def save(self, *args, update_fields=None, **kwargs):
        if update_fields is None or 'description' in update_fields:
            render_results = compile_translated_markdown_fields(self, self.conference, 'description')
            store_relationships(self.conference, self, render_results)

        if not self.slug:
            self.generate_slug()

        return super().save(*args, update_fields=update_fields, **kwargs)

    @classmethod
    def from_dict(
        cls, data: dict, conference: Conference, assembly: Assembly = None, existing=None, pop_used_keys: bool = False, allow_backend_roomtypes: bool = False
    ):
        """
        Loads an Room instance from the given dictionary.
        An existing event can be provided which's data is overwritten (in parts).

        :param data: a dictionary containing Room attributes' names as key (and their values)
        :param assembly: an Assembly instance to which this Room will belong
        :param existing: an instance of Room (or None to get a new one)
        :type existing: Room
        :param pop_used_keys: Remove 'used' keys from the provided data. This can be used to detect additional/errornous fields.
        :param allow_backend_roomtypes: flag whether a backend-based roomtype can be set

        :returns: a new Room or the existing one with fields updated from the given data
        :rtype: Room
        """
        assert isinstance(data, dict), 'Data must be a dictionary.'
        assert existing is None or isinstance(existing, cls), 'Existing instance must be an Room.'

        if existing:
            obj = existing
            if assembly:
                assert obj.assembly == assembly, "Existing room's assembly does not match given one."
                assert obj.conference == assembly.conference, "Existing room's conference does not match the given assembly's one."
            if conference:
                assert obj.conference == conference, "Existing room's conference does not match given one."
        else:
            obj = cls(assembly=assembly, conference=conference)
        assert obj.conference_id is not None

        direct_fields = ['name', 'slug', 'description', 'blocked', 'capacity']
        integer_fields = ['capacity']
        bool_fields = ['blocked']
        for fld in direct_fields:
            if fld in data:
                value = data[fld] if not pop_used_keys else data.pop(fld)
                if fld in integer_fields:
                    value = int(value) if value else None
                elif fld in bool_fields:
                    value = str2bool(value)
                setattr(obj, fld, value)

        if 'room_type' in data:
            v = data['room_type']
            if not allow_backend_roomtypes and v in cls.BACKEND_ROOMTYPES:
                raise ValueError('Room type cannot be set to a backend one.')

            obj.room_type = v
            if pop_used_keys:
                del data['room_type']

        # guess default room_type, if necessary
        if obj.room_type == '':
            obj.room_type = Room.RoomType.STAGE if obj.capacity is None or obj.capacity >= 50 else Room.RoomType.WORKSHOP

        obj.clean()
        return obj

    def get_current_event(self):
        now = timezone.now()

        try:
            event = self.events.conference_accessible(self.conference).get(schedule_start__lte=now, schedule_end__gte=now)
            return event
        except ObjectDoesNotExist:
            return None

    def get_next_event(self):
        now = timezone.now()

        try:
            event = self.events.conference_accessible(self.conference).filter(schedule_start__gt=now).order_by('schedule_start').first()
            return event
        except ObjectDoesNotExist:
            return None


class RoomShare(models.Model):
    class Meta:
        verbose_name = _('RoomShare')
        verbose_name_plural = _('RoomShares')

    id = models.UUIDField(primary_key=True, default=uuid.uuid4, editable=False)
    room = models.ForeignKey(Room, related_name='shares', on_delete=models.CASCADE)

    enabled = models.BooleanField(
        default=False,
        help_text=_('RoomShare__enabled__help'),
        verbose_name=_('RoomShare__enabled'),
    )

    all_self_organized_sessions = models.BooleanField(
        default=False,
        help_text=_('RoomShare__all_self_organized_sessions__help'),
        verbose_name=_('RoomShare__all_self_organized_sessions'),
    )
    all_assemblies = models.BooleanField(
        default=False,
        help_text=_('RoomShare_all_assemblies__help'),
        verbose_name=_('RoomShare_all_assemblies'),
    )
    single_assembly = models.ForeignKey(
        Assembly,
        on_delete=models.SET_NULL,
        blank=True,
        null=True,
        help_text=_('RoomShare__single_assembly__help'),
        verbose_name=_('RoomShare__single_assembly'),
    )

    timeframe = DateTimeRangeField(
        blank=True,
        null=True,
        help_text=_('RoomShare__time_frame__help'),
        verbose_name=_('RoomShare__time_frame'),
    )

    def __str__(self):
        result = '' if self.enabled else 'DISABLED: '
        result += self.room.slug + ' ->'
        if self.all_assemblies:
            result += ' all_assemblies'
        elif self.single_assembly:
            result += f' assembly({self.single_assembly.slug})'
        if self.all_self_organized_sessions:
            result += ' self_organized_sessions'
        if self.timeframe:
            result += f' [{self.timeframe}]'
        return result

    def clean(self):
        errors = {}

        # all and a specific assembly at the same? doesn't make sense
        if self.all_assemblies and self.single_assembly:
            msg = _('RoomShare__all_and_single_assembly_fail')
            errors['all_assemblies'] = msg
            errors['single_assembly'] = msg

        # no share target selected at all?
        if not (self.all_self_organized_sessions or self.all_assemblies or self.single_assembly):
            msg = _('RoomShare__at_least_one_option')
            errors['all_self_organized_sessions'] = msg
            errors['all_assemblies'] = msg
            errors['single_assembly'] = msg

        if errors:
            raise ValidationError(errors)


class RoomLinkManager(ConferenceManagerMixin['RoomLink']):
    # TODO: Add public filter when switching to querysets
    staff_permissions = ['core.assembly_team']
    conference_filter = 'room__assembly__conference'
    assembly_filter = 'room__assembly'

    def apply_manage_filter(self, queryset: 'QuerySet[RoomLink]', member: ConferenceMember) -> 'QuerySet[RoomLink]':
        return queryset.filter(Q(assembly__members__member=member, assembly__members__can_manage_assembly=True))


class RoomLink(models.Model):
    class LinkType(models.TextChoices):
        WEBSITE = 'website', _('RoomLink__type-website')
        CHAT = 'chat', _('RoomLink__type-chat')
        BBB = 'bbb', _('RoomLink__type-bbb')
        JITSI = 'jitsi', _('RoomLink__type-jitsi')
        PAD = 'pad', _('RoomLink__type-pad')
        MEDIA_CCC_DE = 'media.ccc.de', _('RoomLink__type-media_ccc_de')
        VIDEO = 'video', _('RoomLink__type-video')
        AUDIO = 'audio', _('RoomLink__type-audio')
        NAV = 'nav', _('RoomLink__type-nav')

    objects = RoomLinkManager()

    URL_LINKTYPES = [LinkType.WEBSITE, LinkType.CHAT, LinkType.BBB, LinkType.JITSI, LinkType.PAD, LinkType.VIDEO, LinkType.AUDIO, LinkType.NAV]
    """All LinkTypes which require the link to be a valid URL. The notable exception is media_ccc_de where only a global id is required."""

    room = models.ForeignKey(Room, related_name='links', on_delete=models.CASCADE)
    name = models.CharField(max_length=200)

    link_type = models.CharField(max_length=20, choices=LinkType.choices, help_text=_('RoomLink__type__help'), verbose_name=_('RoomLink__type'))
    link = models.CharField(max_length=255)

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self._resolved_url = None

    def _cache_resolved_url_if_necessary(self):
        if self._resolved_url:
            return
        self._resolved_url = resolve_link(self.link)

    @property
    def resolved_url(self) -> str:
        self._cache_resolved_url_if_necessary()
        return self._resolved_url[1]

    @property
    def resolved_url_internal(self) -> bool:
        self._cache_resolved_url_if_necessary()
        return self._resolved_url[0] == 'internal'

    def clean(self):
        if self.link_type in RoomLink.URL_LINKTYPES:
            if not resolve_internal_url(self.link, fallback_as_is=False):
                raise ValidationError({'link': _('RoomLink__link__must_be_url')})

    def to_dict(self):
        return {
            'type': self.link_type,
            'name': self.name,
            'uri': self.link,
            'url': resolve_internal_url(self.link),
        }

    def __str__(self):
        return self.name
