from django.db import models
from django.utils.translation import gettext_lazy as _


class BackendMixin(models.Model):
    @classmethod
    def admin_url(cls, pk):
        from django.urls import reverse

        return reverse(f'admin:{cls._meta.app_label}_{cls._meta.model_name}_change', args=[pk])

    class Meta:
        abstract = True

    class BackendStatus(models.TextChoices):
        NONE = '', _('Room__backend_status-none')
        """
        Not applicable: This room has no backend service.
        """

        NEW = 'new', _('Room__backend_status-new')
        """
        The room is new and the backend has not begun setting it up yet.
        """

        SETUP = 'setup', _('Room__backend_status-setup')
        """
        The backend is currently being configured / set up.
        """

        ERROR = 'error', _('Room__backend_status-error')
        """
        The backend is having a problem with this room.
        """

        ACTIVE = 'active', _('Room__backend_status-active')
        """
        The backend is active, the room can be used.
        """

        INACTIVE = 'inactive', _('Room__backend_status-inactive')
        """
        The backend is inactive, the room is not available.
        """

        FULL = 'full', _('Room__backend_status-full')
        """
        The backend is overcrowded: the room is present but no new guests can be accommodated.
        """

    backend_link = models.URLField(blank=True, null=True, help_text=_('Room__backend_link__help'), verbose_name=_('Room__backend_link'))
    """Reference to integration/source system. For e.g. WorkAdventure this is the collection JSON URL."""

    backend_link_branch = models.CharField(
        max_length=50,
        blank=True,
        null=True,
        help_text=_('Room__backend_link_branch__help'),
        verbose_name=_('Room__backend_link_branch'),
    )
    """Additional reference to integration/source system, i.e. a git branch name."""

    backend_flags = models.JSONField(blank=True, null=True, help_text=_('Room__backend_flags__help'), verbose_name=_('Room__backend_flags'))
    """Arbitrary data which can be used to integration component."""

    backend_data = models.JSONField(blank=True, null=True, help_text=_('Room__backend_data__help'), verbose_name=_('Room__backend_data'))
    """Arbitrary data necessary for room setup in the backend - used by the actual integration component (not for users!!!)."""

    director_data = models.JSONField(blank=True, null=True, help_text=_('Room__director_data__help'), verbose_name=_('Room__director_data'))
    """Arbitrary data necessary for room operation in the backend - used by the actual integration component (not for users!!!)."""

    backend_status = models.CharField(
        max_length=20,
        blank=True,
        choices=BackendStatus.choices,
        default=BackendStatus.NONE,
        help_text=_('Room__backend_status__help'),
        verbose_name=_('Room__backend_status'),
    )
    """The backend's status information."""

    capacity = models.PositiveIntegerField(blank=True, null=True, help_text=_('Room__capacity__help'), verbose_name=_('Room__capacity'))
    """A room's capacity, i.e. maximum of participants."""

    reserve_capacity = models.PositiveIntegerField(
        default=20, help_text=_('Room__reserve_capacity__help'), verbose_name=_('Room__reserve_capacity'), editable=False
    )
    """Number of slots the deployment infrastructure should keep available for this room"""

    occupants = models.PositiveIntegerField(blank=True, null=True, help_text=_('Room__occupants__help'), verbose_name=_('Room__occupants'), editable=False)
    """The (estimated) number of occupants in the room as reported by the backend."""
