from typing import TYPE_CHECKING, TypeIs
from uuid import uuid4

from django.contrib.auth.models import Group
from django.contrib.contenttypes.fields import GenericRelation
from django.db import models
from django.utils.translation import gettext_lazy as _
from rules.contrib.models import RulesModelBase, RulesModelMixin
from rules.predicates import is_superuser, predicate

from core.fields import ConferenceReference
from core.markdown import compile_translated_markdown_fields, store_relationships
from core.models.activitylog import ActivityLogEntry, ActivityLogMixin
from core.models.invitation import Invitation
from core.predicates import is_conference_staff

if TYPE_CHECKING:  # pragma: no cover
    from core.models import PlatformUser


@predicate
def is_team_member(user: 'PlatformUser', team: 'Team | None' = None) -> bool:
    if team is None:  # pragma: no cover
        return False
    return user.is_authenticated and team.members.filter(user=user).exists()


@predicate
def is_team_manager(user: 'PlatformUser', team: 'Team | None' = None) -> bool:
    if team is None:  # pragma: no cover
        return False
    return user.is_authenticated and team.members.filter(user=user, can_manage=True).exists()


@predicate
def is_staff_team(user: 'PlatformUser', team: 'Team | None' = None) -> bool:
    if team is None:  # pragma: no cover
        return False
    return team.visibility == Team.TeamVisibility.STAFF


@predicate
def is_public_team(user: 'PlatformUser', team: 'Team | None' = None) -> bool:
    # If team is None, we are checking for the permission to view the list of teams, everybody can see it.
    if team is None:
        return True
    return team.visibility == Team.TeamVisibility.PUBLIC


class Team(RulesModelMixin, ActivityLogMixin, Group, metaclass=RulesModelBase):
    class TeamVisibility(models.TextChoices):
        PUBLIC = 'public', _('Team__visibility__Public')
        STAFF = 'staff', _('Team__visibility__Staff')
        HIDDEN = 'hidden', _('Team__visibility__Hidden')

    class Meta:
        rules_permissions = {
            'view': is_team_member | is_public_team | is_superuser | (is_conference_staff & is_staff_team),
            'view_details': is_team_member | is_superuser,
            'add': is_superuser,
            'change': is_team_manager | is_superuser,
            'change_permissions': is_superuser,
            'delete': is_superuser,
            'can_join': (is_conference_staff & is_staff_team) | is_public_team | is_superuser,
        }
        indexes = [
            models.Index(fields=['uuid']),
        ]
        verbose_name = _('Team')
        verbose_name_plural = _('Teams')

    uuid = models.UUIDField(unique=True, default=uuid4, editable=False)
    conference = ConferenceReference(
        related_name='Teams',
        verbose_name=_('Team__conference'),
    )
    description = models.TextField(
        blank=True,
        verbose_name=_('Team__description'),
        help_text=_('Team__description__help'),
    )
    description_html = models.TextField(
        blank=True,
        null=True,
        verbose_name=_('Team__description_html'),
    )
    visibility = models.CharField(
        default=TeamVisibility.STAFF,
        verbose_name=_('Team__visibility'),
        help_text=_('Team__visibility__help'),
        choices=TeamVisibility.choices,
    )

    sent_invitations = GenericRelation(
        Invitation,
        content_type_field='requester_type',
        object_id_field='requester_id',
        related_query_name='assemblies_inviters',
    )
    received_invitations = GenericRelation(
        Invitation,
        content_type_field='requested_type',
        object_id_field='requested_id',
        related_query_name='assemblies_invited',
    )

    @classmethod
    def type_is(cls, obj: object) -> TypeIs['Team']:
        return isinstance(obj, cls)

    def save(self, *args, update_fields=None, **kwargs):
        if update_fields is None or 'description' in update_fields:  # pragma: no branch
            render_results = compile_translated_markdown_fields(self, self.conference, 'description')
            store_relationships(self.conference, self, render_results)

        return super().save(*args, update_fields=update_fields, **kwargs)

    def get_activity_log_kind_for_user(self, user: 'PlatformUser') -> ActivityLogEntry.Kind:
        if self.members.filter(user=user).exists():
            return ActivityLogEntry.Kind.ENTITY
        if user.is_superuser:
            return ActivityLogEntry.Kind.STAFF
        return ActivityLogEntry.Kind.OTHER
