import logging
import re
from contextlib import suppress
from pathlib import Path
from random import choices
from string import ascii_lowercase, digits
from typing import TYPE_CHECKING, Any, TypeIs
from uuid import uuid4

from timezone_field import TimeZoneField
from urllib3.util import parse_url

from django.conf import settings
from django.contrib.auth.models import AbstractUser, UserManager
from django.contrib.contenttypes.fields import GenericForeignKey
from django.contrib.contenttypes.models import ContentType
from django.contrib.postgres import fields as pg_fields
from django.core.exceptions import ObjectDoesNotExist, ValidationError
from django.core.mail import get_connection
from django.core.mail.message import (
    EmailMultiAlternatives,
)
from django.core.validators import validate_email
from django.db import models
from django.db.models import Q
from django.utils.functional import cached_property
from django.utils.text import slugify
from django.utils.timezone import now
from django.utils.translation import gettext_lazy as _

from core.fields import ConferenceReference
from core.utils import download_from_url
from core.validators import ImageDimensionValidator

if TYPE_CHECKING:
    from core.models import Assembly, Conference, ConferenceMember

logger = logging.getLogger(__name__)


def get_user_avatar_filename(instance: 'PlatformUser', filename: Path | str):
    ext = str(filename).rsplit('.', maxsplit=1)[-1]
    return Path('avatars').joinpath(f'{instance.id}.{ext}')


def get_user_avatar_help_text() -> str:
    return _('PlatformUser__avatar__help %(min_width)d, %(min_height)d, %(max_width)d, %(max_height)d') % {
        'min_width': settings.USER_AVATAR_WIDTH_MINIMUM,
        'min_height': settings.USER_AVATAR_HEIGHT_MINIMUM,
        'max_width': settings.USER_AVATAR_WIDTH_MAXIMUM,
        'max_height': settings.USER_AVATAR_HEIGHT_MAXIMUM,
    }


class PlatformUserManager(UserManager):
    def get_by_natural_key(self, username: str) -> Any:
        return self.get(**{f'{self.model.USERNAME_FIELD}__iexact': username})

    def has_talk_on_conference(self, conf: 'Conference'):
        from .events import Event, EventParticipant  # late import to mitigate circular dependency

        qs = self.get_queryset()

        events = Event.objects.filter(conference=conf) if conf else Event.objects.all()
        event_owners = events.values_list('owner_id', flat=True).distinct()

        participants = EventParticipant.objects.filter(event__conference=conf) if conf else EventParticipant.objects.all()
        participants = participants.filter(role=EventParticipant.Role.SPEAKER)
        event_speakers = participants.values_list('participant_id', flat=True).distinct()

        qs = qs.filter(Q(pk__in=event_owners) | Q(pk__in=event_speakers))
        return qs


class PlatformUser(AbstractUser):
    class Type(models.TextChoices):
        HUMAN = ('human', _('PlatformUser__type-human'))
        """A human user which registered itself."""
        SPEAKER = ('speaker', _('PlatformUser__type-speaker'))
        """A human user which was created automatically."""
        SERVICE = ('service', _('PlatformUser__type-service'))
        """A technical user representing a service connected to the hub."""
        ASSEMBLY = ('assembly', _('PlatformUser__type-assembly'))
        """A technical user representing an assembly, used e.g. for API and SSO purposes."""
        BOT = ('bot', _('PlatformUser__type-bot'))
        """A technical user taking part in on-platform activity."""

    class Theme(models.TextChoices):
        DARK = ('dark', _('PlatformUser__theme-dark'))
        LIGHT = ('light', _('PlatformUser__theme-light'))

    PERSON_TYPES = [Type.HUMAN, Type.SPEAKER]
    """User types which can be shown as persons on e.g. the frontend."""

    INTERACTIVE_TYPES = [Type.HUMAN, Type.BOT]
    """User types which can be interactive (and can thus have an avatar and/or modified by e.g. the Engelsystem)."""

    objects = PlatformUserManager()

    user_type = models.CharField(
        max_length=10, choices=Type.choices, default=Type.HUMAN, help_text=_('PlatformUser__type__help'), verbose_name=_('PlatformUser__type')
    )

    uuid = models.UUIDField(default=uuid4, unique=True)
    slug = models.SlugField(blank=False, unique=True, max_length=150)
    display_name = models.CharField(max_length=150, blank=True, help_text=_('PlatformUser__display_name__help'), verbose_name=_('PlatformUser__display_name'))

    # self portrayal
    show_name = models.BooleanField(null=True, help_text=_('PlatformUser__show_name__help'), verbose_name=_('PlatformUser__show_name'))
    avatar = models.ImageField(
        blank=True,
        null=True,
        help_text=get_user_avatar_help_text,
        upload_to=get_user_avatar_filename,
        verbose_name=_('PlatformUser__avatar'),
        validators=[
            ImageDimensionValidator(
                min_size=(
                    settings.USER_AVATAR_WIDTH_MINIMUM,
                    settings.USER_AVATAR_HEIGHT_MINIMUM,
                ),
                max_size=(
                    settings.USER_AVATAR_WIDTH_MAXIMUM,
                    settings.USER_AVATAR_HEIGHT_MAXIMUM,
                ),
                square=True,
            ),
        ],
    )
    avatar_url = models.URLField(blank=True, null=True, help_text=_('PlatformUser__avatar_url__help'), verbose_name=_('PlatformUser__avatar_url'))
    pronouns = models.CharField(max_length=50, blank=True, default='', help_text=_('PlatformUser__pronouns__help'), verbose_name=_('PlatformUser__pronouns'))

    status = models.CharField(max_length=50, blank=True, null=True, help_text=_('PlatformUser__status__help'), verbose_name=_('PlatformUser__status'))
    status_public = models.BooleanField(null=True, help_text=_('PlatformUser__status_public__help'), verbose_name=_('PlatformUser__status_public'))
    timezone = TimeZoneField(
        help_text=_('PlatformUser__timezone__help'),
        verbose_name=_('PlatformUser__timezone'),
        default='Europe/Berlin',
        choices_display='WITH_GMT_OFFSET',
    )

    # Accessibility Options
    no_animations = models.BooleanField(default=False, help_text=_('PlatformUser__no_animations__help'), verbose_name=_('PlatformUser__no_animations'))
    colorblind = models.BooleanField(default=False, help_text=_('PlatformUser__colorblind__help'), verbose_name=_('PlatformUser__colorblind'))
    high_contrast = models.BooleanField(default=False, help_text=_('PlatformUser__high_contrast__help'), verbose_name=_('PlatformUser__high_contrast'))
    tag_ignorelist = pg_fields.ArrayField(
        models.SlugField(), blank=True, default=list, help_text=_('PlatformUser__tag_ignore_list__help'), verbose_name=_('PlatformUser__tag_ignore_list')
    )
    theme = models.CharField(
        max_length=50, choices=Theme.choices, default=Theme.DARK, verbose_name=_('PlatformUser__theme'), help_text=_('PlatformUser__theme__help')
    )

    # Disturbance Settings
    receive_dms = models.BooleanField(default=True, help_text=_('PlatformUser__receive_dms__help'), verbose_name=_('PlatformUser__receive_dms'))
    receive_dm_images = models.BooleanField(
        default=False, help_text=_('PlatformUser__receive_dm_images__help'), verbose_name=_('PlatformUser__receive_dm_images')
    )
    receive_audio = models.BooleanField(default=True, help_text=_('PlatformUser__receive_audio__help'), verbose_name=_('PlatformUser__receive_audio'))
    receive_video = models.BooleanField(default=True, help_text=_('PlatformUser__receive_video__help'), verbose_name=_('PlatformUser__receive_video'))

    # Administrative
    shadow_banned = models.BooleanField(default=False, help_text=_('PlatformUser__shadow_banned__help'), verbose_name=_('PlatformUser__shadow_banned'))
    admin_notification = models.TextField(
        default='', blank=True, help_text=_('PlatformUser__admin_notification__help'), verbose_name=_('PlatformUser__admin_notification')
    )

    audio_muted = models.BooleanField(default=False)
    audio_volume = models.FloatField(blank=True, null=True)

    # privacy settings
    autoaccept_contacts = models.BooleanField(
        default=False, help_text=_('PlatformUser__auto_accept_contacts__help'), verbose_name=_('PlatformUser__auto_accept_contacts')
    )

    is_searchable = models.BooleanField(default=False, help_text=_('PlatformUser__is_searchable__help'), verbose_name=_('PlatformUser__is_searchable'))

    # menu settings
    menu = models.JSONField(blank=True, default=dict, help_text=_('PlatformUser__menu__help'), verbose_name=_('PlatformUser__menu'))

    # security settings
    allow_reset_non_primary = models.BooleanField(
        default=False, help_text=_('PlatformUser__allow_reset_non_primary__help'), verbose_name=_('PlatformUser__allow_reset_non_primary')
    )

    @classmethod
    def get_user_flags(cls, user):
        return {
            'colorblind': user.colorblind,
            'high_contrast': user.high_contrast,
            'no_animations': user.no_animations,
        }

    def __str__(self):
        return self.username

    @classmethod
    def get_anonymous_user(cls):
        logger.trace('Creating anonymous user.')

        class AnonUser:
            def __init__(self):
                for x in cls._meta.fields:
                    setattr(self, x.name, x.default if x.default != models.fields.NOT_PROVIDED else None)

                self.show_name = False
                self.receive_dms = False
                self.username = ''

            is_authenticated = False

            def save(self, *args, **kwargs):
                raise TypeError('Cannot save an anonymous user.')

        return AnonUser()

    @classmethod
    def type_is(cls, obj: object) -> TypeIs['PlatformUser']:
        return isinstance(obj, cls)

    def get_all_verified_addresses(self):
        """Return a list of all verified addresses (for legal logging purposes, e.g. hangar creation)."""
        return self.communication_channels.filter(is_verified=True).values_list('address', flat=True)

    def get_avatar_color(self, conference):
        try:
            cm = conference.users.get(user=self)

            if cm.active_angel:
                return 'FFFF00'

            return 'FFFFFF'

        except ObjectDoesNotExist:
            return None

    def get_absolute_url(self):
        from core.templatetags.hub_absolute import hub_absolute  # pylint: disable=import-outside-toplevel

        return hub_absolute('plainui:user', user_slug=self.slug, i18n=settings.ARCHIVE_MODE)

    def is_in_conference(self, conference):
        return conference.users.filter(user=self).exists()

    def is_conference_staff(self, conference):
        return conference.users.filter(user=self, is_staff=True).exists()

    @property
    def is_person(self):
        return self.user_type in self.PERSON_TYPES

    @cached_property
    def name(self) -> str:
        """
        Shows the display name with username if they differ.

        This is used to make the name available in a way matching other models (e.g. for invitations).

        ReturnsL
            str: The users display name.
        """
        name = self.display_name
        if self.first_name or self.last_name:
            name += f' <{self.username}>'
        return name

    def get_display_name(self):
        result = ''
        if self.first_name:
            result += self.first_name
        if self.last_name:
            result += f' {self.last_name}'
        result = result.strip()
        if not result:
            result = self.username
        if self.pronouns:
            result += f' ({self.pronouns})'
        return result

    @property
    def guardians(self):
        return self.contacts.filter(is_guardian=True).annotate(pk=models.F('contact__pk'), username=models.F('contact__username')).values('pk', 'username')

    def get_staticpage_groups(self, conference):
        """Fetches the ConferenceMember's static_page_groups."""
        try:
            cm = conference.users.get(user=self)
            return cm.static_page_groups
        except ObjectDoesNotExist:
            return None

    @classmethod
    def from_dict(
        cls,
        data: dict,
        conference: 'Conference',
        assembly: 'Assembly' = None,
        existing=None,
        pop_used_keys: bool = False,
    ):
        """
        Loads an PlatformUser instance from the given dictionary - used by the Schedule import mechanism.
        An existing user can be provided which's data is overwritten (in parts).
        However, for sanity reasons, this will only operate on user_type=SPEAKER.

        :param data: a dictionary containing PlatformUser attributes' names as key (and their values)
        :param conference: the Conference instance to which this PlatformUser shall be joined
        :param assembly: ignored
        :param existing: an instance of PlatformUser (or None to get a new one)
        :type existing: PlatformUser
        :param pop_used_keys: Remove 'used' keys from the provided data. This can be used to detect additional/errornous fields.

        :returns: a new user or the existing one with fields updated from the given data, in both cases joined to the conference
        :rtype: PlatformUser
        """
        assert isinstance(data, dict), 'Data must be a dictionary.'
        if not existing:
            raise NotImplementedError('Creating a PlatformUser with .from_dict() is not supported.')

        obj = existing

        # join the new user into the conference
        cm, created = conference.users.get_or_create(user=obj)
        if existing.user_type == cls.Type.SPEAKER or created:
            # only set the description/bio upon creation for real users (imported speakers are updated every time)
            cm.description = data.get('biography', '')
        cm.save()

        # add all known addresses as (unverified and non-public) communication channels
        for addr in data.get('addresses', []):
            obj.communication_channels.get_or_create(
                channel=UserCommunicationChannel.Channel.MAIL,
                address=addr,
                defaults={
                    'is_verified': False,
                    'show_public': False,
                },
            )

        # load avatar, if an URL is given (and has not changed since last time)
        if avatar_url := data.get('avatar'):
            if obj.avatar_url != avatar_url:
                try:
                    obj.load_avatar_from_url(avatar_url)
                except Exception:
                    # just log the error, if the speaker has no avatar (yet), we don't care too much
                    logger.exception('Failed to download avatar for (new) speaker user %s: %s', obj.pk, avatar_url)

    def __create_slug(self, extension=''):
        """
        recursive function to get a free slug based on the username and an optional extension string.
        """
        user_slug = slugify(self.username + extension)

        # If a user has a username that consists only of non slug characters, create a random slug.
        if user_slug == '':
            user_slug = ''.join(choices(ascii_lowercase + digits, k=6))

        if PlatformUser.objects.exclude(username=self.username).filter(slug=user_slug).exists():
            # Slug already in use.
            extension = str(uuid4())[13:18]
            return self.__create_slug(user_slug + extension)
        else:
            return user_slug

    def generate_slug(self):
        self.slug = self.__create_slug()

    def save(self, *args, update_fields=None, **kwargs):
        if not self.slug:
            self.generate_slug()

        # update the display name
        self.display_name = self.get_display_name()
        with suppress(AttributeError):
            del self.name  # delete the cached property to force a refresh

        return super().save(*args, update_fields=update_fields, **kwargs)

    def has_conference_staff_permission(self, conference: 'Conference', *perms: str, require_all: bool = False):
        """Returns True if this user is a staff member of the given conference and has,
        depending upon require_all, either at least one of the supplied permissions or all of them.

        Args:
            conference (Conference): The Conference instance to check against.
            *perms (str): The permissions to check for.
            require_all (bool, optional): Defines, if all permission are required or just one. Defaults to False.

        Returns:
            bool: True if the user has all permissions.
        """

        if conference is None:
            # no conference given means we're not staff there
            return False

        try:
            cm: ConferenceMember = conference.users.get(user=self)
            return cm.has_perms(*perms, require_all=require_all, require_staff=True)
        except ObjectDoesNotExist:
            return False

    def notify_user(self, subject, message_text, message_html=None, details_link=None, use_all_verified_channels=False, from_email=None, reply_to=None):
        """
        Sends the given subject and message via all communication channels the user has configured to be used for notifications.
        The message must be provided in text form, the HTML version is optional.
        Please note that some channels might only receive the subject due to message length restrictions.
        The details_link attribute holds an URL which should be linked with the message if the full message cannot be delivered.

        For example, an email would contain subject+message whereas a Twitter DM would include subject and the details_link but _not_ the message.
        """
        success, fail = 0, 0
        extra_filters = {} if use_all_verified_channels else {'use_for_notifications': True}
        for channel in self.communication_channels.filter(is_verified=True, **extra_filters):
            if channel.send(
                subject, message_text, message_html, details_link, allow_unverified=use_all_verified_channels, from_email=from_email, reply_to=reply_to
            ):
                success += 1
            else:
                fail += 1
        return fail == 0

    def has_verified_mail_addresses(self):
        return self.communication_channels.filter(
            channel=UserCommunicationChannel.Channel.MAIL,
            is_verified=True,
        ).exists()

    def get_verified_mail_addresses(self):
        return list(
            self.communication_channels.filter(
                channel=UserCommunicationChannel.Channel.MAIL,
                is_verified=True,
            ).values_list('address', flat=True)
        )

    def get_primary_mail(self):
        return (
            self.communication_channels.filter(
                channel=UserCommunicationChannel.Channel.MAIL,
                is_verified=True,
                is_primary=True,
            )
            .values_list('address', flat=True)
            .first()
        )

    def load_avatar_from_url(self, url: str, save: bool = True):
        logger.debug('Loading avatar for user %s (%s) from URL: %s', self.username, self.pk, url)

        filename, data = download_from_url(url)
        target_filename = get_user_avatar_filename(self, filename)

        logger.debug('Storing avatar for user %s (%s): %s', self.username, self.pk, target_filename)
        self.avatar.save(target_filename, data)
        self.avatar_url = url
        if save:
            self.save(update_fields=['avatar', 'avatar_url'])
        logger.debug('Updated avatar for user %s (%s) as "%s" from URL: %s', self.username, self.pk, target_filename, url)


class UserContact(models.Model):
    user = models.ForeignKey(PlatformUser, related_name='contacts', on_delete=models.CASCADE)
    contact = models.ForeignKey(PlatformUser, related_name='+', on_delete=models.CASCADE)

    pending = models.BooleanField(default=True, help_text=_('UserContact__pending__help'), verbose_name=_('UserContact__pending'))

    share_status = models.BooleanField(null=True, help_text=_('UserContact__share_status__help'), verbose_name=_('UserContact__share_status'))

    receive_dms = models.BooleanField(default=True, help_text=_('UserContact__receive_dms__help'), verbose_name=_('UserContact__receive_dms'))
    receive_dm_images = models.BooleanField(
        default=False, help_text=_('UserContact__receive_dm_images__help'), verbose_name=_('UserContact__receive_dm_images')
    )

    is_guardian = models.BooleanField(default=False, help_text=_('UserContact__is_guardian__help'), verbose_name=_('UserContact__is_guardian'))

    @cached_property
    def reverse_contact(self):
        return UserContact.objects.get(user=self.contact, contact=self.user)

    def reverse_contact_share_status(self, default=None):
        try:
            return self.reverse_contact.share_status
        except UserContact.DoesNotExist:
            return default


_RE_MATRIX_ROOM_URI = re.compile(r'^(?P<sigil>[!#])(?P<localpart>[A-Za-z0-9][^:]*):(?P<domain>.+)$')
_RE_MATRIX_TO_ROOM_LINK = re.compile(r'^https://matrix\.to/#/(%23|!)[A-Za-z0-9].+?%3A[^/]+$')
_RE_TELEPHONE = re.compile(r'^(00|\+)\d+[\d -]*$')
_RE_VANITY = re.compile(r'^(00|\+)49[- ]?(180|[789]00)[- ]?[A-Za-z0-9 -]+$')


class UserCommunicationChannel(models.Model):
    class Meta:
        constraints = [models.UniqueConstraint(fields=['user', 'channel', 'is_primary'], condition=Q(is_primary=True), name='unique_channel_primary')]

    class Channel(models.TextChoices):
        MAIL = 'mail', _('UserCommunicationChannel__channel-mail')
        XMPP = 'xmpp', _('UserCommunicationChannel__channel-xmpp')
        IRC = 'irc', _('UserCommunicationChannel__channel-irc')
        DECT = 'dect', _('UserCommunicationChannel__channel-dect')
        PHONE = 'phone', _('UserCommunicationChannel__channel-phone')
        MATRIX = 'matrix', _('UserCommunicationChannel__channel-matrix')
        ACTIVITYPUB = 'activitypub', _('UserCommunicationChannel__channel-activitypub')

    user = models.ForeignKey(PlatformUser, related_name='communication_channels', on_delete=models.CASCADE)

    channel = models.CharField(
        max_length=20, choices=Channel.choices, help_text=_('UserCommunicationChannel__channel__help'), verbose_name=_('UserCommunicationChannel__channel')
    )
    address = models.CharField(max_length=255, help_text=_('UserCommunicationChannel__address__help'), verbose_name=_('UserCommunicationChannel__address'))

    caption = models.CharField(
        max_length=200, blank=True, null=True, help_text=_('UserCommunicationChannel__caption__help'), verbose_name=_('UserCommunicationChannel__caption')
    )

    is_verified = models.BooleanField(
        default=False, help_text=_('UserCommunicationChannel__is_verified__help'), verbose_name=_('UserCommunicationChannel__is_verified')
    )
    is_primary = models.BooleanField(
        default=False, help_text=_('UserCommunicationChannel__is_primary__help'), verbose_name=_('UserCommunicationChannel__is_primary')
    )

    use_for_notifications = models.BooleanField(
        default=False, help_text=_('UserCommunicationChannel__use_for_notifications__help'), verbose_name=_('UserCommunicationChannel__use_for_notifications')
    )
    show_public = models.BooleanField(
        default=False, help_text=_('UserCommunicationChannel__show_public__help'), verbose_name=_('UserCommunicationChannel__show_public')
    )

    _logger = logging.getLogger(__name__)

    def __init__(self, *args: Any, **kwargs: Any) -> None:
        super().__init__(*args, **kwargs)
        # Save the original address for comparison in the save method
        self.__original_address = self.address

    def __str__(self):
        return f'{self.channel}={self.address}' + (' (unverified)' if not self.is_verified else '')

    @staticmethod
    def validate_email(address):
        try:
            validate_email(address)
        except ValidationError as err:
            raise ValidationError({'address': str(err)})

    @staticmethod
    def validate_matrix_room_url(address):
        if _RE_MATRIX_ROOM_URI.match(address) is None and _RE_MATRIX_TO_ROOM_LINK.match(address) is None:
            raise ValidationError({'address': 'expected Matrix Room URI or matrix.to link'})

    @staticmethod
    def validate_phone_number(address):
        if _RE_TELEPHONE.match(address) is None and _RE_VANITY.match(address) is None:
            raise ValidationError({'address': 'expected an international telephone number (starting with + or 00)'})

    @staticmethod
    def validate_activitypub_url(address):
        url = parse_url(address)
        if url.scheme not in ['http', 'https']:
            raise ValidationError({'address': 'Expected https:// URL.'})
        if url.host is None or url.path is None:
            raise ValidationError({'address': 'Expected valid URL.'})

    @staticmethod
    def validate_dect(address):
        try:
            if len(str(address)) > 10:
                raise ValidationError({'address': 'Dect may not be longer than 10 chars.'})
            if address[0] == '0':
                raise ValidationError({'address': 'Dect must not start with a 0.'})
            int(address)
        except ValueError:
            raise ValidationError({'address': 'Dect must be a number.'})

    @property
    def can_notify(self):
        """Signals whether this channel can be used for notifications."""
        return self.channel in [self.Channel.MAIL, self.Channel.MATRIX]

    def clean(self):
        super().clean()

        match self.channel:
            case self.Channel.MAIL:
                self.validate_email(self.address)
            case self.Channel.MATRIX:
                self.validate_matrix_room_url(self.address)
            case self.Channel.DECT:
                self.validate_dect(self.address)
                if self.use_for_notifications:
                    raise ValidationError({'use_for_notifications': _('UserCommunicationChannel__cannot_notify__dect')})
            case self.Channel.PHONE:
                self.validate_phone_number(self.address)
                if self.use_for_notifications:
                    raise ValidationError({'use_for_notifications': _('UserCommunicationChannel__cannot_notify__phone')})
            case self.Channel.ACTIVITYPUB:
                self.validate_activitypub_url(self.address)

        # TODO: verify the other channel types for correct syntax as well

        if self.use_for_notifications:
            raise ValidationError({'use_for_notifications': _('UserCommunicationChannel__cannot_notify')})

        if self.is_primary:
            if not self.is_verified:
                raise ValidationError({'is_primary': _('UserCommunicationChannel__primary_must_be_verified')})

    def save(self, *args, **kwargs):
        other_primary = UserCommunicationChannel.objects.exclude(pk=self.id).filter(is_primary=True, channel=self.channel, user=self.user)
        if self.is_primary and other_primary:
            other_primary.update(is_primary=False)
        elif not other_primary:
            self.is_primary = True
        if self.__original_address != self.address:
            self.is_verified = False
        super().save(*args, **kwargs)

    def send_mail(
        self,
        subject,
        message,
        from_email,
        recipient_list,
        fail_silently=False,
        auth_user=None,
        auth_password=None,
        connection=None,
        html_message=None,
        reply_to=None,
    ):
        """
        Easy wrapper for sending a single message to a recipient list. All members
        of the recipient list will see the other recipients in the 'To' field.

        If from_email is None, use the DEFAULT_FROM_EMAIL setting.
        If auth_user is None, use the EMAIL_HOST_USER setting.
        If auth_password is None, use the EMAIL_HOST_PASSWORD setting.
        If reply_to is None, use the MAIL_REPLY_TO setting.

        Note: The API for this method is frozen. New code wanting to extend the
        functionality should use the EmailMessage class directly.
        """
        connection = connection or get_connection(
            username=auth_user,
            password=auth_password,
            fail_silently=fail_silently,
        )
        reply_to = reply_to if reply_to is not None else settings.MAIL_REPLY_TO
        mail = EmailMultiAlternatives(subject, message, from_email, recipient_list, connection=connection, reply_to=reply_to)
        if html_message:
            mail.attach_alternative(html_message, 'text/html')

        return mail.send()

    def send(self, subject, message_text, message_html=None, send_to_unverified=False, allow_unverified=False, from_email=None, reply_to=None):
        if not allow_unverified and not self.use_for_notifications:
            raise Exception('Channel not marked to be used for notifications.')
        if not self.is_verified and not send_to_unverified:
            raise Exception('Cannot send to unverified channel.')

        if self.channel == self.Channel.MAIL:
            try:
                result = self.send_mail(
                    recipient_list=[self.address],
                    from_email=from_email,
                    subject=subject,
                    message=message_text,
                    html_message=message_html,
                    reply_to=reply_to,
                )
                self._logger.info(
                    'Sent %(result)s mail to "%(receiver)s" <%(address)s> with subject "%(subject)s".',
                    {'result': result, 'receiver': self.user.username, 'address': self.address, 'subject': subject},
                )
                return True

            except Exception as err:
                self._logger.exception(
                    'Failed to send mail to "%(receiver)s" <%(address)s> with subject "%(subject)s": %(error)s',
                    {'receiver': self.user.username, 'address': self.address, 'subject': subject, 'error': err},
                )
                return False

        raise NotImplementedError(f"Don't know how to send to a channel of type {self.channel}")


class UserTimelineEntry(models.Model):
    id = models.UUIDField(default=uuid4, primary_key=True)

    user = models.ForeignKey(PlatformUser, related_name='timeline', on_delete=models.CASCADE)
    conference = ConferenceReference(related_name='+')

    timestamp = models.DateTimeField(default=now, help_text=_('UserTimelineEntry__timestamp__help'), verbose_name=_('UserTimelineEntry__timestamp'))

    target_type = models.ForeignKey(ContentType, on_delete=models.CASCADE)
    target_id = models.UUIDField()
    target = GenericForeignKey('target_type', 'target_id')

    is_bookmark = models.BooleanField(default=False, help_text=_('UserTimelineEntry__is_bookmark__help'), verbose_name=_('UserTimelineEntry__is_bookmark'))
    """Indicates a timeline entry 'saved for later'."""

    comment = models.TextField(blank=True, null=True, help_text=_('UserTimelineEntry__comment__help'), verbose_name=_('UserTimelineEntry__comment'))
