from django.conf import settings
from django.conf.urls.i18n import i18n_patterns
from django.contrib import admin
from django.http import HttpRequest
from django.template.defaulttags import register
from django.urls import NoReverseMatch, get_script_prefix, include, path, reverse, set_script_prefix
from django.utils import translation

PREFIXES = {
    'admin': settings.ADMIN_BASE_URL,
    'api': settings.API_BASE_URL,
    'backoffice': settings.BACKOFFICE_BASE_URL,
    'plainui': settings.PLAINUI_BASE_URL,
}


class PlainUIUrlPatterns:
    def __init__(self) -> None:
        self.urlpatterns = i18n_patterns(
            path('', include('plainui.urls', namespace='plainui')),
            prefix_default_language=True,
        )


class UrlPatternWrapper:
    def __init__(self, namespace, urls) -> None:
        self.urlpatterns = [path('', include(urls, namespace=namespace))]


URL_CONFIGS = {
    'plain': {
        'admin': admin.site.urls,
        'api': UrlPatternWrapper('api', 'api.urls'),
        'backoffice': UrlPatternWrapper('backoffice', 'backoffice.urls'),
        'plainui': UrlPatternWrapper('plainui', 'plainui.urls'),
    },
    'i18n': {
        'admin': admin.site.urls,
        'api': UrlPatternWrapper('api', 'api.urls'),
        'backoffice': UrlPatternWrapper('backoffice', 'backoffice.urls'),
        'plainui': PlainUIUrlPatterns(),
    },
}


@register.simple_tag
def hub_absolute(url: str, *args, i18n: bool = True, lang: str | None = None, query_string: str = '', **kwargs) -> str:
    """
    Generates an absolute URL for the given view name, with optional internationalization and query string.

    ### Args:
        - url (str):\
            The view name in the format 'namespace:viewname'.
        - *args:
          Positional arguments to pass to the URL resolver.
        - i18n (bool, optional): Whether to use internationalization. Defaults to True.
        - lang (str | None, optional): Language code to switch to. Defaults to None.
        - query_string (str, optional): Query string to append to the URL. Defaults to ''.
        - **kwargs: Keyword arguments to pass to the URL resolver.

    ### Returns:
        - str: The generated absolute URL.

    ### Raises:
        NoReverseMatch: If the URL cannot be reversed due to a missing namespace.
    """

    try:
        app, _ = url.split(':')
    except ValueError as e:
        raise NoReverseMatch('No reverse found, missing namespace') from e
    thread_prefix = get_script_prefix()
    set_script_prefix(PREFIXES[app])
    cur_language = translation.get_language()
    try:
        if lang:
            # switch language if requested
            translation.activate(lang)
        reverse_url = reverse(url, URL_CONFIGS['i18n' if i18n else 'plain'][app], current_app=app, args=args, kwargs=kwargs)
    finally:
        translation.activate(cur_language)
    set_script_prefix(thread_prefix)
    return reverse_url if not query_string else f'{reverse_url}?{query_string}'


@register.simple_tag
def hub_absolute_self(request: HttpRequest, i18n: bool = True, lang: str | None = None):
    """
    Generates an absolute URL for the current requests url.

    ### Args:
        - request (HttpRequest): The HTTP request object.
        - i18n (bool, optional): Flag to include internationalization in the URL. Defaults to True.
        - lang (str | None, optional): Language code to include in the URL. Defaults to None.

    ### Returns:
        - str: The absolute URL for the current requests url.
    """

    url = f'{request.resolver_match.app_name}:{request.resolver_match.url_name}'
    query_string = request.META.get('QUERY_STRING')

    return hub_absolute(url, i18n=i18n, lang=lang, query_string=query_string, *request.resolver_match.args, **request.resolver_match.kwargs)
