import logging
from datetime import datetime, timezone

from django.core.exceptions import ValidationError
from django.test import TestCase

from core.models.assemblies import Assembly
from core.models.conference import Conference, ConferenceMember
from core.models.users import PlatformUser
from core.models.voucher import Voucher


class VoucherTests(TestCase):
    def setUp(self):
        self.conference = Conference(
            slug='foo',
            name='Foo Conference',
            start=datetime(2023, 12, 1, 9, 0, 0, tzinfo=timezone.utc),
            end=datetime(2023, 12, 31, 15, 00, 00, tzinfo=timezone.utc),
            require_ticket=True,
        )
        self.conference.save()

        self.user1 = PlatformUser(username='bernd')
        self.user1.save()
        ConferenceMember(conference=self.conference, user=self.user1).save()

        self.user2 = PlatformUser(username='chili')
        self.user2.save()
        ConferenceMember(conference=self.conference, user=self.user2).save()

        self.guest = PlatformUser(username='elton')
        self.guest.save()

        self.assembly = Assembly(conference=self.conference, slug='fnord', name='Fnord Assembly')
        self.assembly.state = Assembly.State.ACCEPTED
        self.assembly.save()

    def testBasics(self):
        v = Voucher(conference=self.conference, name='test')

        # check defaults
        self.assertEqual(v.assignment, Voucher.Assignment.MANUAL)
        self.assertEqual(v.data, Voucher.Data.TEXT)

        # check that target has no default and therefore an immediate save() fails
        with self.assertRaises(ValidationError):
            v.save()
        v.target = Voucher.Target.ASSEMBLY
        v.save()

        # create an entry, verify that it's not assigned anywhere
        entry = v.entries.create(content='abc')  # type: VoucherEntry
        self.assertIsNone(entry.assigned)
        self.assertIsNone(entry.assigned_assembly)
        self.assertIsNone(entry.assigned_user)

        # create a second voucher and an entry for it, too
        v2 = self.conference.vouchers.create(name='uv', target=Voucher.Target.USER)  # type: Voucher
        entry2 = v2.entries.create(content='abc')  # type: VoucherEntry
        self.assertIsNone(entry2.assigned)
        self.assertIsNone(entry2.assigned_assembly)
        self.assertIsNone(entry2.assigned_user)

    def testVoucherVisibility(self):
        v1 = self.conference.vouchers.create(name='public', assignment=Voucher.Assignment.MANUAL, target=Voucher.Target.ASSEMBLY, show_always=True)  # type: Voucher # noqa: E501
        v2 = self.conference.vouchers.create(name='private', assignment=Voucher.Assignment.MANUAL, target=Voucher.Target.ASSEMBLY, show_always=False)  # type: Voucher # noqa: E501

        # establish baseline, enabled is set to False by default
        self.assertEqual(0, Voucher.objects.available_for_conference(self.conference).count())

        # enable both and check both are available
        v1.enabled = True
        v1.save()
        v2.enabled = True
        v2.save()
        self.assertEqual(2, Voucher.objects.available_for_conference(self.conference).count())

        # check for our assembly, it should show only the public one as available (but unassigned)
        self.assertEqual(1, Voucher.objects.for_assembly(self.assembly).count())
        self.assertEqual(0, self.assembly.get_voucher_count())

        # create an entry on the private voucher and assign it to our test assembly
        v2.entries.create(content='test').assign(self.assembly)
        self.assertEqual(2, Voucher.objects.for_assembly(self.assembly).count())
        self.assertEqual(1, self.assembly.get_voucher_count())

        # create more entries, should not affect the above values
        v1.entries.create(content='more1')
        v2.entries.create(content='more2')
        self.assertEqual(2, Voucher.objects.for_assembly(self.assembly).count())
        self.assertEqual(1, self.assembly.get_voucher_count())

        # assign the public voucher to the assembly, it should now have two assigned
        v1.entries.create(content='foo').assign(self.assembly)
        self.assertEqual(2, Voucher.objects.for_assembly(self.assembly).count())
        self.assertEqual(2, self.assembly.get_voucher_count())

    def testManualAssignment(self):
        # setup test objects
        v = self.conference.vouchers.create(name='test', assignment=Voucher.Assignment.MANUAL, target=Voucher.Target.ASSEMBLY, enabled=True)  # type: Voucher
        entry = v.entries.create(content='abc')  # type: VoucherEntry
        v2 = self.conference.vouchers.create(name='uv', assignment=Voucher.Assignment.MANUAL, target=Voucher.Target.USER, enabled=True)  # type: Voucher
        entry2 = v2.entries.create(content='abc')  # type: VoucherEntry

        # try assigning entries to wrong targets
        with self.assertRaises(ValueError, msg='Entry for voucher with target=Assembly accepted an user.'):
            entry.assign(self.user1)
        with self.assertRaises(ValueError, msg='Entry for voucher with target=Assembly accepted an guest.'):
            entry.assign(self.guest)
        with self.assertRaises(ValueError, msg='Entry for voucher with target=User accepted an assembly.'):
            entry2.assign(self.assembly)

        # try assigning an entry to a guest
        with self.assertRaises(ValueError, msg='Could assign an entry for voucher with target=User to someone not being a member of the conference?!'):
            entry2.assign(self.guest)

        # assign entries to targets and ensure resulting data
        entry.assign(self.assembly)
        entry2.assign(self.user1)

        # assigning them again should not be possible
        with self.assertRaises(ValidationError):
            entry2.assign(self.user2)

    def doAutoAssignment(self, expected_warnings=0, conference=None, assemblies=None) -> int:
        if expected_warnings == 0:
            with self.assertNoLogs(level=logging.WARNING):
                total = Voucher.do_auto_assignments(conference=conference, assemblies=assemblies)

        else:
            with self.assertLogs(level=logging.WARNING) as log:
                total = Voucher.do_auto_assignments(conference=conference, assemblies=assemblies)
            self.assertEqual(expected_warnings, len(log.records))

        return total

    def testAutomaticAssignment(self):
        # setup test objects
        v = self.conference.vouchers.create(name='test', assignment=Voucher.Assignment.ON_PUBLIC, target=Voucher.Target.ASSEMBLY, enabled=True)  # type: Voucher
        entry = v.entries.create(content='abc')  # type: VoucherEntry

        # trigger auto-assignment (would be run by e.g. cron)
        self.assertEqual(1, self.doAutoAssignment())

        # check that the entry has been assigned
        entry.refresh_from_db()
        self.assertEqual(entry.target, self.assembly)

        # setup a second assembly which is not 'public' yet
        a2 = self.conference.assemblies.create(slug='second', name='test2', state=Assembly.State.REGISTERED)

        # try auto-assignment again, should have no effect (assignment=ON_PUBLIC, only one public assembly which is already assigned)
        self.assertEqual(0, self.doAutoAssignment())

        # set assembly public, should trigger a warning as we don't have a free entry available
        a2.state = Assembly.State.ACCEPTED
        a2.save()
        self.assertEqual(0, self.doAutoAssignment(expected_warnings=1))

        # add another entry and we should be fine
        entry2 = v.entries.create(content='def')  # type: VoucherEntry
        self.assertEqual(1, self.doAutoAssignment())
        entry2.refresh_from_db()
        self.assertEqual(entry2.target, a2)
