__all__ = (
    'AwardBadgeView',
    'BadgeDetailView',
    'BadgeListView',
    'BadgeSettingsView',
    'ManageBadgeView',
    'ManageBadgesView',
    'RedeemTokenDetailView',
)

from typing import Any

from django_ratelimit.decorators import ratelimit

from django.conf import settings
from django.contrib import messages
from django.db.models import QuerySet
from django.http import HttpRequest, HttpResponse, HttpResponseRedirect
from django.shortcuts import redirect
from django.urls import reverse, reverse_lazy
from django.utils.decorators import method_decorator
from django.utils.translation import gettext_lazy as _
from django.views.generic.base import TemplateView
from django.views.generic.detail import DetailView
from django.views.generic.edit import FormView, UpdateView

from core.models import Assembly, Badge, BadgeToken, UserBadge
from core.models.badges import TokenInvalid
from core.models.conference import ConferenceMember
from core.views.list_views import FilteredListView

from plainui.forms import (
    BadgeDefaultsEditForm,
    ManageBadgeForm,
    RedeemBadgeForm,
)
from plainui.views.utils import ConferenceRequiredMixin


class BadgeListView(ConferenceRequiredMixin, FilteredListView):
    model = Badge
    template_name = 'plainui/badges/list.html.j2'
    paginate_by = 50
    filters = {
        'category': ('category__slug', 'str'),
        'assembly': ('issuing_assembly__slug', 'str'),
    }
    ordering = ['name']

    def get_context_data(self, **kwargs: Any) -> dict[str, Any]:
        user = self.request.user
        owned = user.badges.all().values_list('badge__pk', flat=True) if user.is_authenticated else []
        categories = self.conf.badge_categories.all()
        return {
            'conf': self.conf,
            'categories': categories,
            'owned_badges': owned,
            'redeem_badge_form': RedeemBadgeForm(),
            **super().get_context_data(**kwargs),
        }

    def get_queryset(self) -> QuerySet[BadgeToken]:
        user = self.request.user
        qs: QuerySet[Badge] = Badge.objects.accessible_by_user(user, self.conf, staff_can_manage=False)
        get_parameters = getattr(self.request, 'GET', {})
        if owned := get_parameters.get('owned'):
            self.params['owned'] = owned
            owned = owned == 'true'
            if not user.is_authenticated:
                messages.error(self.request, 'BadgeList__not_authenticated')
                return Badge.objects.none() if owned else qs
            qs = qs.filter(users__user=user, users__accepted_by_user=True) if owned else qs.exclude(users__user=user, users__accepted_by_user=True)
        return super().get_queryset(qs).distinct()


class BadgeDetailView(ConferenceRequiredMixin, DetailView):
    model = Badge
    template_name = 'plainui/badges/detail.html.j2'
    context_object_name = 'badge'

    def get_queryset(self):
        user = self.request.user
        qs: QuerySet[Badge] = Badge.objects.accessible_by_user(user, self.conf, staff_can_manage=False).distinct()
        return qs

    def get_context_data(self, **kwargs: Any) -> dict[str, Any]:
        user = self.request.user
        badge: Badge = self.get_object()
        manageable = badge.issuing_assembly.user_can_manage(user) if user.is_authenticated else False
        return {'conf': self.conf, 'manageable': manageable, **super().get_context_data(**kwargs)}


class AwardBadgeView(ConferenceRequiredMixin, DetailView):
    model = Badge
    template_name = 'plainui/badges/award.html.j2'

    def get(self, request: HttpRequest, *args: Any, **kwargs: Any) -> HttpResponse:
        badge = self.get_object()
        if not badge.users.filter(user=request.user.id).exists():
            return HttpResponseRedirect(reverse('plainui:badge', kwargs={'pk': badge.pk}))
        return super().get(request, *args, **kwargs)

    def get_context_data(self, **kwargs: Any) -> dict[str, Any]:
        return {'conf': self.conf, **super().get_context_data(**kwargs)}


class RedeemTokenDetailView(ConferenceRequiredMixin, DetailView):
    model = BadgeToken
    template_name = 'plainui/badges/token_detail.html.j2'
    context_object_name = 'token'

    def get_context_data(self, **kwargs: Any) -> dict[str, Any]:
        # Regenerate the qr code if needed
        if self.object.qr_full is None:
            self.object.save()
        return {
            **super().get_context_data(**kwargs),
            'token_svg': self.object.qr_full.read().decode('utf-8') if self.object.qr_full else None,
            'conf': self.conf,
        }

    def get_queryset(self) -> QuerySet[BadgeToken]:
        manageable = Assembly.objects.manageable_by_user(user=self.request.user, conference=self.conf, staff_can_manage=False)
        qs = super().get_queryset()
        qs = qs.filter(badge__issuing_assembly__in=manageable)
        return qs


@method_decorator(ratelimit(group='redeem_badge', key='user', rate=settings.BADGE_RATE_LIMIT), name='dispatch')
class RedeemBadgeView(ConferenceRequiredMixin, FormView):
    require_user = True
    template_name = 'plainui/badges/manage.html.j2'
    form_class = RedeemBadgeForm
    external_context = None

    def __init__(self, **kwargs: Any) -> None:
        self.external_context = kwargs.pop('external_context', {})
        super().__init__(**kwargs)

    def get_form_kwargs(self) -> dict[str, Any]:
        kwargs = super().get_form_kwargs()
        kwargs['conference'] = self.conf
        return kwargs

    def form_valid(self, form: Any) -> HttpResponse:
        token = form.cleaned_data['token']
        try:
            user_badge, created = UserBadge.objects.redeem_badge(user=self.request.user, token=token)
            if created:
                messages.success(self.request, _('Badge__awarded %(badge)s') % {'badge': user_badge.badge})
            else:
                messages.info(self.request, _('Badge__already_owned %(badge)s') % {'badge': user_badge.badge})
        except TokenInvalid as e:
            messages.error(self.request, e.msg)
            return HttpResponseRedirect(reverse('plainui:manage_badges'))
        return HttpResponseRedirect(reverse('plainui:badge_award', kwargs={'pk': user_badge.badge.pk}))

    def get_context_data(self, **kwargs: Any) -> dict[str, Any]:
        context = super().get_context_data(**kwargs)
        context['conf'] = self.conf
        context.update(self.external_context)
        context['redeem_badge_form'] = context.pop('form')
        context['badge_defaults_form'] = BadgeDefaultsEditForm(instance=ConferenceMember.get_member(user=self.request.user, conference=self.conf))
        return context

    def get(self, request: HttpRequest, *args: str, **kwargs: Any) -> HttpResponse:
        token = request.GET.get('redeem_token')
        try:
            self.form = RedeemBadgeForm({'token': token, 'purpose': 'redeem_token'}, conference=self.conf)
            if self.form.is_valid():
                return self.form_valid(self.form)
            return self.form_invalid(self.form)
        except BadgeToken.DoesNotExist:
            messages.error(self.request, _('BadgeToken__does_not_exist %(token)s') % {'token': token})
        return super().get(request, *args, **kwargs)


class BadgeSettingsView(ConferenceRequiredMixin, UpdateView):
    model = ConferenceMember
    require_user = True
    success_url = reverse_lazy('plainui:manage_badges')
    form_class = BadgeDefaultsEditForm
    external_context = None

    def __init__(self, **kwargs: Any) -> None:
        self.external_context = kwargs.pop('external_context', {})
        super().__init__(**kwargs)

    def get_object(self) -> QuerySet[Any]:
        return ConferenceMember.get_member(user=self.request.user, conference=self.conf)

    def form_valid(self, form: Any) -> HttpResponse:
        form.instance.default_badge_visibility = form.cleaned_data['default_badge_visibility']
        form.instance.save()

        messages.success(self.request, _('Badge__settings_success'))
        return super().form_valid(form)

    def form_invalid(self, form):
        messages.error(self.request, _('Badge__settings_error'))
        return redirect(self.get_success_url())

    def get_context_data(self, **kwargs: Any) -> dict[str, Any]:
        context = super().get_context_data(**kwargs)
        context['conf'] = self.conf
        context.update(self.external_context)
        return context

    def get(self, request: HttpRequest, *args: Any, **kwargs: Any) -> HttpResponse:
        return HttpResponseRedirect(reverse('plainui:manage_badges'))


class ManageBadgesView(ConferenceRequiredMixin, TemplateView):
    require_user = True
    template_name = 'plainui/badges/manage.html.j2'
    mode = None

    def get_context_data(self, *args, **kwargs):
        context = super().get_context_data(*args, **kwargs)
        if hasattr(self, 'conf'):
            context['conf'] = self.conf

        user = self.request.user
        context['badges_accepted'] = []
        context['badges_not_accepted'] = []
        for badge in user.badges.all() if user.is_authenticated else []:
            if badge.accepted_by_user:
                context['badges_accepted'].append(ManageBadgeForm(instance=badge))
            else:
                context['badges_not_accepted'].append(ManageBadgeForm(instance=badge))
        context['redeem_badge_form'] = RedeemBadgeForm()
        context['disable_share'] = True
        if hasattr(self, 'conf'):
            context['badge_defaults_form'] = BadgeDefaultsEditForm(instance=ConferenceMember.get_member(user=self.request.user, conference=self.conf))
        return context

    def dispatch(self, request, *args, **kwargs):
        if (hasattr(request, 'GET') and request.GET.get('redeem_token')) or (hasattr(request, 'GET') and request.POST.get('purpose', None) == 'redeem_token'):
            return RedeemBadgeView.as_view(external_context=self.get_context_data())(request, *args, **kwargs)
        return super().dispatch(request, *args, **kwargs)


class ManageBadgeView(ConferenceRequiredMixin, UpdateView):
    model = UserBadge
    form_class = ManageBadgeForm
    require_user = True
    mode = None
    success_url = reverse_lazy('plainui:manage_badges')

    def get_queryset(self):
        return UserBadge.objects.filter(user=self.request.user)

    def get_object(self, queryset: QuerySet[UserBadge] | None = None) -> UserBadge:
        if queryset is None:
            queryset = self.get_queryset()
        return queryset.get(badge__id=self.kwargs.get('pk'))

    def dispatch(self, request, *args, **kwargs):
        if (hasattr(request, 'GET') and request.GET.get('redeem_token')) or (hasattr(request, 'GET') and request.POST.get('purpose', None) == 'redeem_token'):
            return RedeemBadgeView.as_view(external_context=self.get_context_data())(request, *args, **kwargs)
        return super().dispatch(request, *args, **kwargs)

    def form_invalid(self, form):
        messages.error(self.request, _('Badge__update_error %(badge)s') % {'badge': form.instance.badge})
        return redirect(self.get_success_url())

    def form_valid(self, form):
        if self.mode == 'accept':
            form.instance.accepted_by_user = True
            form.instance.save()
            return HttpResponseRedirect(reverse('plainui:badge_award', kwargs={'pk': form.instance.badge.pk}))
        elif self.mode == 'revoke':
            form.instance.accepted_by_user = False
            messages.success(self.request, _('Badge__revoked %(badge)s') % {'badge': form.instance.badge})
        elif self.mode == 'update':
            messages.success(self.request, _('Badge__update__successful %(badge)s') % {'badge': form.instance.badge})
        return super().form_valid(form)
