__all__ = (
    'BoardEntryCreateView',
    'BoardEntryDeleteView',
    'BoardEntryEditView',
    'BoardEntryView',
    'BoardPrivateView',
    'BoardView',
)


from django_ratelimit.decorators import ratelimit

from django.contrib import messages
from django.db.models.query import Q, QuerySet
from django.urls import reverse, reverse_lazy
from django.utils.decorators import method_decorator
from django.utils.translation import gettext
from django.views.generic import DeleteView, DetailView, ListView
from django.views.generic.edit import CreateView, FormMixin, UpdateView

from core.models import (
    BulletinBoardEntry,
)

from plainui.forms import (
    BulletinBoardEntryForm,
)
from plainui.views.utils import ConferenceRequiredMixin


class BoardView(ConferenceRequiredMixin, ListView):
    model = BulletinBoardEntry
    template_name = 'plainui/board.html.j2'
    context_object_name = 'board'
    paginate_by = 20

    def get_queryset(self) -> QuerySet[BulletinBoardEntry]:
        """
        Return a queryset of all entries that are public and not hidden, or are owned by the current user.
        """
        user = self.request.user
        # General board view only shows public entries
        qs = super().get_queryset().select_related('owner').filter(conference=self.conf, is_public=True).order_by('-timestamp')
        if user.is_authenticated:
            # If user is logged in, show their own entries and entries from other users that are not hidden
            # with this a shadow-banned user can still see their own entries, but other users can't see them
            qs = qs.filter(Q(owner__shadow_banned=False, hidden=False) | Q(owner=user))
        else:
            # If user is not logged in, only show entries from non-shadow-banned users that are not hidden
            qs = qs.filter(owner__shadow_banned=False, hidden=False)
        return qs

    def get_context_data(self, **kwargs):
        return {
            **super().get_context_data(**kwargs),
            'conf': self.conf,
            'current_user': self.request.user.id if self.request.user.is_authenticated else 0,
            'report_info': {'enabled': True},
            'private': False,
        }


class BoardPrivateView(BoardView):
    require_user = True

    def get_queryset(self) -> QuerySet[BulletinBoardEntry]:
        return BulletinBoardEntry.objects.filter(conference=self.conf, owner=self.request.user).order_by('-timestamp')

    def get_context_data(self, **kwargs):
        return {**super().get_context_data(**kwargs), 'private': True}


class BoardEntryView(ConferenceRequiredMixin, DetailView):
    model = BulletinBoardEntry
    context_object_name = 'entry'
    template_name = 'plainui/board_entry.html.j2'

    def get_queryset(self) -> QuerySet[BulletinBoardEntry]:
        """
        Return a queryset of all entries that are public and not hidden, or are owned by the current user.
        """
        # For details see BoardView.get_queryset
        user = self.request.user
        qs = super().get_queryset().select_related('owner').filter(conference=self.conf, is_public=True).order_by('-timestamp')
        qs = qs.filter()
        if user.is_authenticated:
            qs = qs.filter(Q(owner__shadow_banned=False, hidden=False) | Q(owner=user))
        else:
            qs = qs.filter(owner__shadow_banned=False, hidden=False)
        return qs

    def get_context_data(self, **kwargs):
        return {
            **super().get_context_data(**kwargs),
            'conf': self.conf,
            'current_user': self.request.user.id if self.request.user.is_authenticated else 0,
            'report_info': {'kind': 'board', 'url': self.kwargs['pk']},
        }


@method_decorator(ratelimit(key='user', rate='3/m', method=ratelimit.UNSAFE, block=False), name='dispatch')
@method_decorator(ratelimit(key='ip', rate='5/m', method=ratelimit.UNSAFE, block=False), name='dispatch')
class BoardEntryFormMixin(ConferenceRequiredMixin, FormMixin):
    model = BulletinBoardEntry
    form_class = BulletinBoardEntryForm
    template_name = 'plainui/board_edit.html.j2'
    require_user = True

    def get_form_kwargs(self):
        return {
            **super().get_form_kwargs(),
            'conference': self.conf,
            'request': self.request,
        }

    def get_context_data(self, **kwargs):
        return {
            **super().get_context_data(**kwargs),
            'conf': self.conf,
            'disable_share': True,
        }

    def get_success_url(self):
        return reverse('plainui:board_entry_edit', kwargs={'pk': self.object.id})

    def form_valid(self, form):
        self.board_entry = form.save(commit=False)
        self.board_entry.hidden = self.request.user.shadow_banned
        self.board_entry.save()

        if 'pk' in self.kwargs:
            messages.success(self.request, gettext('Bulletin Board Entry updated.'))
        else:
            messages.success(self.request, gettext('Bulletin Board Entry created.'))
        return super().form_valid(form)


class BoardEntryCreateView(BoardEntryFormMixin, CreateView):
    pass


class BoardEntryEditView(BoardEntryFormMixin, UpdateView):
    def get_queryset(self) -> QuerySet[BulletinBoardEntry]:
        return super().get_queryset().filter(owner=self.request.user)


class BoardEntryDeleteView(ConferenceRequiredMixin, DeleteView):
    model = BulletinBoardEntry
    require_user = True
    success_url = reverse_lazy('plainui:board_private')

    def get_queryset(self) -> QuerySet[BulletinBoardEntry]:
        return super().get_queryset().filter(owner=self.request.user)

    def form_valid(self, form):
        response = super().form_valid(form)
        messages.success(self.request, gettext('Bulletin Board Entry deleted.'))
        return response
