__all__ = (
    'StaticPageDiffView',
    'StaticPageEditView',
    'StaticPageGlobalHistoryView',
    'StaticPageHistoryView',
    'StaticPageLockKeepalive',
    'StaticPageRedirectToStart',
    'StaticPageView',
)

from datetime import timedelta
from typing import Any

from django_ratelimit.decorators import ratelimit

from django.conf import settings
from django.contrib import messages
from django.contrib.contenttypes.models import ContentType
from django.db import transaction
from django.http import Http404, HttpResponse, HttpResponseForbidden
from django.shortcuts import redirect
from django.template.response import TemplateResponse
from django.urls import reverse
from django.utils.decorators import method_decorator
from django.utils.translation import get_language, gettext
from django.views import View
from django.views.generic.base import TemplateView

from core.markdown import refresh_linking_markdown, render_markdown
from core.models import (
    Lock,
    StaticPage,
    StaticPageRevision,
)

from plainui.forms import (
    StaticPageBodyForm,
)
from plainui.utils import StaticPageDiff
from plainui.views.utils import ConferenceRequiredMixin


class StaticPageView(ConferenceRequiredMixin, TemplateView):
    template_name = 'plainui/static_page.html.j2'
    require_conference_member = False

    @transaction.atomic
    def get(self, request, *args, page_slug, **kwargs):
        self.static_page = StaticPage.objects.conference_accessible(conference=self.conf, language=get_language()).filter(slug=page_slug).first()

        if 'release' in request.GET:
            lock = Lock.objects.filter(
                content_type=ContentType.objects.get_for_model(StaticPage), object_id=self.static_page.pk, pk=request.GET['release']
            ).first()
            if lock:
                lock.delete()

        if self.static_page:
            if self.static_page.privacy == StaticPage.Privacy.CONFERENCE and not self.has_ticket(self.request):
                if not self.request.user.is_authenticated:
                    return self.handle_no_permission()
                else:
                    messages.error(self.request, gettext('You need an active Ticket to access this Page!'))
                    return redirect(reverse('plainui:redeem_token'))

        return super().get(request, *args, page_slug=page_slug, **kwargs)

    def get_context_data(self, page_slug, **kwargs):
        context = super().get_context_data(**kwargs)
        context['conf'] = self.conf
        context['page_slug'] = page_slug
        context['page'] = static_page = self.static_page
        context['page_can_edit'] = can_edit = static_page is not None
        context['page_no_permission'] = False

        if not static_page:
            context['page_can_create'] = False

            if self.request.user.is_authenticated:
                # check if user can create static page
                static_page, _page_exists = StaticPage.objects.get_editable_page(self.request.user, self.conf, get_language(), page_slug, check=True)

                # get_editable_page always returns a StaticPage instance if the user is allowed to create one, even if it doesn't exist in the database
                context['page_can_create'] = static_page is not None
            return context

        if static_page.privacy == StaticPage.Privacy.PERM:
            if not self.request.user.is_authenticated or not self.request.user.has_conference_staff_permission(self.conf, 'core.static_pages'):
                context['page_no_permission'] = True
                context['page_can_edit'] = False
                context['page'] = None
                return context

        context['page_can_create'] = False
        revision = self.request.GET.get('rev')
        if revision:
            try:
                static_page_revision = static_page.revisions.get(revision=revision)
                # TODO: this might be a security issue IF evil code got saved in an revision and sanitization was switched off later?
                page_body = render_markdown(
                    self.conf,
                    static_page_revision.body,
                    allow_embedded_html=not static_page.remove_html,
                    sanitize_html=static_page.sanitize_html,
                )
                if can_edit:
                    context['edit_url'] = reverse('plainui:static_page_edit', kwargs={'page_slug': page_slug}) + ('?rev=' + revision if revision else '')
                else:
                    context['edit_url'] = None
            except StaticPageRevision.DoesNotExist:
                context['revision_not_found'] = True
                page_body = static_page.body_html
        else:
            page_body = static_page.body_html
            context['edit_url'] = reverse('plainui:static_page_edit', kwargs={'page_slug': page_slug}) if can_edit else None

        context['page_body'] = page_body
        context['report_info'] = {'lookup_key': static_page.slug}
        return context


class StaticPageEditView(ConferenceRequiredMixin, TemplateView):
    template_name = 'plainui/static_page_edit.html.j2'
    require_user = True
    require_conference_member = False

    @transaction.atomic
    def get(self, request, *args, page_slug, **kwargs):
        LOCK_TIMEOUT = settings.STATIC_PAGE_LOCK_TIMEOUT
        static_page, page_exists = StaticPage.objects.get_editable_page(request.user, self.conf, get_language(), page_slug, check=True)
        writeable = True
        if static_page is None:  # get_editable_page returned permission denied
            static_page = StaticPage.objects.conference_accessible(conference=self.conf, language=get_language()).filter(slug=page_slug).first()
            writeable = False
        elif not page_exists:
            static_page.title = page_slug
            if not settings.STATIC_PAGE_LOCALIZED_BY_DEFAULT:
                static_page.language = None

        if not static_page:
            # the page does not exist and the user does not have the permission to create it
            messages.error(request, gettext('You do not have the required permissions to create this page.'))
            return redirect(reverse('plainui:static_page', kwargs={'page_slug': page_slug}))

        if static_page.privacy == StaticPage.Privacy.CONFERENCE and not self.has_ticket(request):
            messages.error(request, gettext('You do not have the required permissions to edit this page.'))
            # TODO: after redeem, redirect to edit view for page_slug
            return redirect(reverse('plainui:redeem_token'))

        if static_page.privacy == StaticPage.Privacy.PERM and not self.request.user.has_conference_staff_permission(self.conf, 'core.static_pages'):
            messages.error(request, gettext('You do not have the required permissions to edit this page.'))
            return redirect(reverse('plainui:static_page', kwargs={'page_slug': page_slug}))

        lock = ''
        if page_exists and writeable:
            lock, created = Lock.objects.select_for_update().get_or_create(
                content_type=ContentType.objects.get_for_model(StaticPage),
                object_id=static_page.pk,
                defaults={
                    'timeout': self.now + timedelta(seconds=LOCK_TIMEOUT),
                    'lock_holder': request.user,
                },
            )
            if created:
                lock = str(lock.pk)
            elif lock.lock_holder == request.user:
                lock.timeout = self.now + timedelta(seconds=LOCK_TIMEOUT)
                lock.save()
                lock = str(lock.pk)
            elif lock.timeout <= self.now:  # lock is expired
                lock.delete()
                lock = Lock(
                    content_type=ContentType.objects.get_for_model(StaticPage),
                    object_id=static_page.pk,
                    lock_holder=request.user,
                    timeout=self.now + timedelta(seconds=LOCK_TIMEOUT),
                )
                lock.save()
                lock = str(lock.pk)
            else:
                writeable = False
                lock = ''

        revision = request.GET.get('rev')
        if revision:
            static_page_revision = static_page.revisions.get(revision=revision)
            instance = static_page_revision
            page_title = static_page_revision.title
            not_latest_revision = static_page_revision.revision != static_page.public_revision
        else:
            instance = static_page
            page_title = static_page.title
            not_latest_revision = False

        form = StaticPageBodyForm(
            initial={
                'title': page_title,
                'body': instance.body,
                'lock': lock,
            }
        )

        if not writeable:
            form.fields['title'].disabled = True
            form.fields['body'].disabled = True

        return TemplateResponse(
            request,
            self.template_name,
            {
                'page': static_page,
                'conf': self.conf,
                'page_exists': page_exists,
                'page_slug': page_slug,
                'static_page': static_page,
                'writeable': writeable,
                'not_latest_revision': not_latest_revision,
                'revision': revision,
                'form': form,
                'disable_share': True,
                'lock_id': lock,
                'lock_endpoint': reverse('plainui:static_page_refresh_lock'),
                'lock_refresh_interval': LOCK_TIMEOUT // 3,
            },
        )

    @method_decorator(ratelimit(key='ip', rate='5/m', method=ratelimit.UNSAFE))
    @method_decorator(ratelimit(key='post:username', rate='5/m', method=ratelimit.UNSAFE))
    @transaction.atomic
    def post(self, request, *args, page_slug, **kwargs):
        static_page, page_exists = StaticPage.objects.get_editable_page(request.user, self.conf, get_language(), page_slug, check=False)
        if not static_page:
            return HttpResponseForbidden()
        elif not page_exists:
            static_page.title = page_slug
            if not settings.STATIC_PAGE_LOCALIZED_BY_DEFAULT:
                static_page.language = None

        if static_page.privacy == StaticPage.Privacy.CONFERENCE and not self.has_ticket(request):
            return redirect(reverse('plainui:static_page', kwargs={'page_slug': page_slug}))

        if static_page.privacy == StaticPage.Privacy.PERM:
            if not self.request.user.has_conference_staff_permission(self.conf, 'core.static_pages'):
                return redirect(reverse('plainui:static_page', kwargs={'page_slug': page_slug}))

        form = StaticPageBodyForm(request.POST)
        preview = request.POST.get('preview')
        revision = request.GET.get('rev')
        try:
            not_latest_revision = revision and int(revision) != static_page.public_revision
        except ValueError:
            not_latest_revision = False

        if not form.is_valid():
            return TemplateResponse(
                request,
                self.template_name,
                {
                    'page': static_page,
                    'conf': self.conf,
                    'page_slug': page_slug,
                    'static_page': static_page,
                    'writeable': True,  # otherwise, `static_page` would be `None` and we'd have raised `PermissionDenied` earlier
                    'not_latest_revision': not_latest_revision,
                    'revision': revision,
                    'form': form,
                    'disable_share': True,
                },
            )

        title = form.cleaned_data['title']
        body = form.cleaned_data['body']

        lock_id = form.cleaned_data['lock']
        lock_error = False
        lock = None
        if page_exists and not preview:
            lock = Lock.objects.filter(pk=lock_id, content_type=ContentType.objects.get_for_model(StaticPage), object_id=static_page.pk).first()
            if not lock or (lock.timeout >= self.now and lock.lock_holder != request.user):
                messages.error(request, gettext('Lock error, could not save page!'))
                lock_error = True

        # refresh lock when previewing
        if not lock_error and page_exists and preview:
            Lock.objects.filter(
                pk=lock_id,
                content_type=ContentType.objects.get_for_model(StaticPage),
                object_id=static_page.pk,
                lock_holder=self.request.user,
            ).update(timeout=self.now + timedelta(seconds=settings.STATIC_PAGE_LOCK_TIMEOUT))

        if lock_error or preview is not None:
            preview_body = render_markdown(
                self.conf,
                body,
                allow_embedded_html=not static_page.remove_html,
                sanitize_html=static_page.sanitize_html,
            )

            return TemplateResponse(
                request,
                self.template_name,
                {
                    'page': static_page,
                    'preview_title': title,
                    'preview_body': preview_body,
                    'conf': self.conf,
                    'page_slug': page_slug,
                    'static_page': static_page,
                    'writeable': True,  # otherwise, `static_page` would be `None` and we'd error out early
                    'not_latest_revision': not_latest_revision,
                    'revision': revision,
                    'form': form,
                    'disable_share': True,
                },
            )

        if not page_exists:
            static_page.save()
            refresh_linking_markdown(self.conf, static_page)
        revision = StaticPageRevision(page=static_page, title=title, author=request.user, body=body)
        revision.save()
        revision.set_public()
        revision.save()

        if not page_exists:
            messages.success(request, gettext('Created Static Page'))
        else:
            messages.success(request, gettext('Updated Static Page'))

        if lock:
            lock.delete()
        return redirect(reverse('plainui:static_page', kwargs={'page_slug': page_slug}))


class StaticPageHistoryView(ConferenceRequiredMixin, TemplateView):
    template_name = 'plainui/static_page_history.html.j2'
    require_user = True

    def get(self, request, *args, page_slug, **kwargs):
        self.static_page = StaticPage.objects.conference_accessible(conference=self.conf, language=get_language()).filter(slug=page_slug).first()
        if not self.static_page:
            return redirect(reverse('plainui:static_page', kwargs={'page_slug': page_slug}))

        if self.static_page.privacy == StaticPage.Privacy.CONFERENCE and not self.has_ticket(self.request):
            return self.handle_no_permission()

        if self.static_page.privacy == StaticPage.Privacy.PERM:
            if not self.request.user.is_authenticated or not self.request.user.has_conference_staff_permission(self.conf, 'core.static_pages'):
                return redirect(reverse('plainui:static_page', kwargs={'page_slug': page_slug}))

        return super().get(request, *args, page_slug=page_slug, **kwargs)

    def get_context_data(self, page_slug, **kwargs) -> dict[str, Any]:
        context = super().get_context_data(**kwargs)
        context['conf'] = self.conf
        context['page_slug'] = page_slug

        context['page'] = self.static_page
        context['history'] = self.static_page.revisions.filter(is_draft=False).select_related('author').order_by('-timestamp', '-pk')
        return context


class StaticPageDiffView(ConferenceRequiredMixin, TemplateView):
    template_name = 'plainui/static_page_diff.html.j2'
    require_user = True

    def get(self, request, *args, page_slug, **kwargs):
        self.static_page = StaticPage.objects.conference_accessible(conference=self.conf, language=get_language()).filter(slug=page_slug).first()
        if not self.static_page:
            return redirect(reverse('plainui:static_page', kwargs={'page_slug': page_slug}))

        if self.static_page.privacy == StaticPage.Privacy.CONFERENCE and not self.has_ticket(self.request):
            return self.handle_no_permission()

        if self.static_page.privacy == StaticPage.Privacy.PERM:
            if not self.request.user.is_authenticated or not self.request.user.has_conference_staff_permission(self.conf, 'core.static_pages'):
                return redirect(reverse('plainui:static_page', kwargs={'page_slug': page_slug}))

        return super().get(request, *args, page_slug=page_slug, **kwargs)

    def get_context_data(self, page_slug, **kwargs) -> dict[str, Any]:
        context = super().get_context_data(**kwargs)
        context['conf'] = self.conf
        context['page_slug'] = page_slug

        context['page'] = self.static_page
        rev1_id = self.request.GET.get('rev')
        rev2_id = self.request.GET.get('rev2')

        if rev1_id is None or rev2_id is None:
            raise Http404('Only one revision given, need two')

        try:
            rev1 = self.static_page.revisions.get(revision=rev1_id)
            rev2 = self.static_page.revisions.get(revision=rev2_id)
        except StaticPageRevision.DoesNotExist:
            raise Http404('One of the selected revisions was not found')

        differ = StaticPageDiff(tabsize=4)
        diff = differ.make_table(rev1.body.splitlines(keepends=True), rev2.body.splitlines(keepends=True), context=True, numlines=3)
        context['rev_id'] = {'rev1': rev1_id, 'rev2': rev2_id}
        context['diff'] = diff
        return context


class StaticPageGlobalHistoryView(ConferenceRequiredMixin, TemplateView):
    template_name = 'plainui/static_page_global_history.html.j2'
    require_user = True

    def get_context_data(self, **kwargs: Any) -> dict[str, Any]:
        PAGE_SIZE = 30
        try:
            page = int(self.request.GET.get('page', '0'))
        except ValueError:
            page = 0

        # TODO tgr respect permissions to list more than publicly-accessible pages? (-> protection)
        static_pages = StaticPage.objects.conference_accessible(self.conf, language=get_language())

        allowed_privacy = [StaticPage.Privacy.NONE]
        if self.has_ticket(self.request):
            allowed_privacy.append(StaticPage.Privacy.CONFERENCE)
        if self.request.user.is_authenticated and self.request.user.has_conference_staff_permission(self.conf, 'core.static_pages'):
            allowed_privacy.append(StaticPage.Privacy.PERM)

        static_pages = static_pages.filter(privacy__in=allowed_privacy)
        history = StaticPageRevision.objects.filter(page__in=static_pages).select_related('page', 'author').order_by('-timestamp', '-pk')

        history_by_day = {}
        for history_entry in history[PAGE_SIZE * page : PAGE_SIZE * page + PAGE_SIZE]:
            day = history_entry.timestamp.strftime('%Y-%m-%d')
            if day not in history_by_day:
                history_by_day[day] = []

            history_by_day[day].append(history_entry)

        context = super().get_context_data(**kwargs)
        context['conf'] = self.conf
        context['history'] = history_by_day
        return context


class StaticPageLockKeepalive(ConferenceRequiredMixin, View):
    def post(self, request):
        LOCK_TIMEOUT = settings.STATIC_PAGE_LOCK_TIMEOUT
        LOCK_MAX_TIMEOUT = settings.STATIC_PAGE_LOCK_MAX_DURATION

        page_slug = request.POST['page_slug']
        lock_id = request.POST['lock_id']

        static_page, page_exists = StaticPage.objects.get_editable_page(self.request.user, self.conf, get_language(), page_slug, check=True)

        if not page_exists:
            return HttpResponse('')

        Lock.objects.filter(
            pk=lock_id,
            content_type=ContentType.objects.get_for_model(StaticPage),
            object_id=static_page.pk,
            lock_holder=self.request.user,
            created_at__gte=self.now - timedelta(seconds=LOCK_MAX_TIMEOUT),
        ).update(timeout=self.now + timedelta(seconds=LOCK_TIMEOUT))

        return HttpResponse('')


class StaticPageRedirectToStart(ConferenceRequiredMixin, View):
    def get(self, request):
        return redirect(reverse('plainui:static_page', kwargs={'page_slug': 'start'}))
