__all__ = (
    'ModifyFavoritesView',
    'ModifyThemeView',
    'ProfileView',
)

from typing import Any

from django.conf import settings
from django.contrib import messages
from django.core.exceptions import SuspiciousOperation
from django.http import HttpResponse, HttpResponseRedirect
from django.shortcuts import redirect
from django.urls import reverse
from django.utils.http import url_has_allowed_host_and_scheme
from django.utils.translation import gettext
from django.views.generic.base import View
from django.views.generic.edit import UpdateView
from modeltranslation.fields import build_localized_fieldname
from modeltranslation.settings import AVAILABLE_LANGUAGES

from core.models import (
    Assembly,
    Event,
    Project,
    UserBadge,
    UserCommunicationChannel,
    UserDereferrerAllowlist,
)
from core.sso import SSO

from plainui.forms import (
    ProfileDescriptionEditForm,
    ProfileEditForm,
    RedeemBadgeForm,
)
from plainui.views.utils import (
    ConferenceRequiredMixin,
    _session_refresh_favorite_assemblies,
    _session_refresh_favorite_events,
    _session_refresh_favorite_projects,
    session_get_favorite_assemblies,
    session_get_favorite_events,
    session_get_favorite_projects,
)


class ProfileView(ConferenceRequiredMixin, UpdateView):
    require_user = True
    only_published = False
    template_name = 'plainui/profile.html.j2'
    form_class = ProfileEditForm

    def get(self, request, *args, **kwargs):
        resp = super().get(request, *args, **kwargs)
        resp.set_cookie(settings.SSO_COOKIE_NAME, SSO.generate_token(self.request.user), domain=settings.SSO_COOKIE_DOMAIN)
        return resp

    def get_object(self, *args, **kwargs):
        return self.request.user

    def get_form(self, form_class=None):
        form1 = super().get_form(form_class)

        form2_kwargs = super().get_form_kwargs()
        if hasattr(self, 'object'):
            # TODO: Update after deciding oh one or more conferences in #648
            cm = self.object.conferences.filter(conference=self.conf).first()
            if cm:
                form2_kwargs['instance'] = cm
            else:
                form2_kwargs.pop('instance')
        form2 = ProfileDescriptionEditForm(**form2_kwargs)
        return form1, form2

    def get_context_data(self, *args, form=None, **kwargs):
        if form:
            form1, form2 = form
        else:
            form1, form2 = self.get_form(self.form_class)
        context = super().get_context_data(*args, form=form1, **kwargs)
        context['conf'] = self.conf
        context['form1'] = form1
        context['form2'] = form2

        user = self.request.user
        owned = user.badges.all().values_list('badge__pk', flat=True) if user.is_authenticated else []

        context['badges'] = UserBadge.objects.filter(user=user, accepted_by_user=True).select_related('badge')
        context['owned_badges'] = owned
        context['amount_badges_not_accepted'] = len(UserBadge.objects.filter(user=user, accepted_by_user=False).select_related('badge'))
        context['is_favorite_events'] = favorite_events = session_get_favorite_events(self.request.session, self.request.user)
        context['my_favorite_events'] = (
            Event.objects.conference_accessible(conference=self.conf)
            .filter(pk__in=favorite_events)
            .filter(schedule_start__isnull=False, schedule_end__isnull=False)
            .order_by('schedule_start')
        )
        context['is_favorite_projects'] = favorite_projects = session_get_favorite_projects(self.request.session, self.request.user)
        context['my_favorite_projects'] = Project.objects.conference_accessible(conference=self.conf).filter(pk__in=favorite_projects).order_by('name')
        context['is_favorite_assemblies'] = favorite_assemblies = session_get_favorite_assemblies(self.request.session, self.request.user)
        context['my_favorite_assemblies'] = Assembly.objects.conference_accessible(conference=self.conf).filter(pk__in=favorite_assemblies).order_by('name')
        context['dereferrer_allowlist'] = UserDereferrerAllowlist.objects.filter(user=self.request.user)
        context['redeem_badge_form'] = RedeemBadgeForm()
        context['disable_share'] = True
        return context

    def get_success_url(self):
        return reverse('plainui:userprofile')

    def post(self, request, *args: Any, **kwargs: Any) -> HttpResponse:
        self.object = self.get_object()
        form1, form2 = self.get_form()
        if form1.is_valid() and form2.is_valid():
            return self.form_valid(form1, form2)
        else:
            return self.form_invalid((form1, form2))

    def form_valid(self, form1, form2):
        form1.instance.pronouns = form1.cleaned_data['pronouns']
        form1.instance.timezone = form1.cleaned_data['timezone']
        form1.instance.save()

        if form1.cleaned_data['dect']:
            UserCommunicationChannel.objects.update_or_create(
                user=self.request.user,
                channel=UserCommunicationChannel.Channel.DECT,
                defaults={'address': form1.cleaned_data['dect']},
            )
        else:
            dect_channel = UserCommunicationChannel.objects.filter(
                user=self.request.user,
                channel=UserCommunicationChannel.Channel.DECT,
            )
            if dect_channel.exists():
                dect_channel.delete()

        # TODO: Update after deciding oh one or more conferences in #648
        cm = form1.instance.conferences.filter(conference=self.conf).first()
        if cm:
            for lang in AVAILABLE_LANGUAGES:
                field_name = build_localized_fieldname('description', lang)
                setattr(cm, field_name, form2.cleaned_data[field_name])
            cm.save(update_fields=['description'])

        form2.save()
        self.request.session['theme'] = form1.instance.theme
        messages.success(self.request, gettext('Updated Profile'))

        return HttpResponseRedirect(self.get_success_url())


class ModifyThemeView(ConferenceRequiredMixin, View):
    def get(self, request, **kwargs):
        return redirect(reverse('plainui:index'))

    def post(self, request):
        redirect_to = request.POST.get('next') or reverse('plainui:index')
        url_is_safe = url_has_allowed_host_and_scheme(
            url=redirect_to,
            allowed_hosts=[self.request.get_host()],
            require_https=self.request.is_secure(),
        )
        if not url_is_safe:
            redirect_to = reverse('plainui:index')

        new_high_contrast = request.POST['high_contrast'] == 'True'
        if self.request.user.is_authenticated:
            self.request.user.high_contrast = new_high_contrast
            self.request.user.save(update_fields=['high_contrast'])

        self.request.session['high_contrast'] = new_high_contrast
        return redirect(redirect_to)


class ModifyFavoritesView(ConferenceRequiredMixin, View):
    require_user = True

    def get(self, request, **kwargs):
        return redirect(reverse('plainui:userprofile'))

    def post(self, request):
        # the implementation here with if exists(): remove() is racy but should be good enough here.
        # if like counts skew too much, we should simply recompute them from scratch.
        fav_id = request.POST['id']
        user = request.user
        match (request.POST['mode'], request.POST['type']):
            case ('add', 'assembly'):
                if Assembly.objects.conference_accessible(conference=self.conf).filter(pk=fav_id).exists():
                    user.favorite_assemblies.add(fav_id)
            case ('remove', 'assembly'):
                user.favorite_assemblies.remove(fav_id)
            case ('add', 'event'):
                if Event.objects.conference_accessible(conference=self.conf).filter(pk=fav_id).exists():
                    user.favorite_events.add(fav_id)
            case ('remove', 'event'):
                user.favorite_events.remove(fav_id)
            case ('add', 'project'):
                if Project.objects.conference_accessible(conference=self.conf).filter(pk=fav_id).exists():
                    user.favorite_projects.add(fav_id)
            case ('remove', 'project'):
                user.favorite_projects.remove(fav_id)
            case _:
                raise SuspiciousOperation('Invalid favorite mode or type requested')

        if request.POST['type'] == 'assembly':
            _session_refresh_favorite_assemblies(self.request.session, self.request.user)
        elif request.POST['type'] == 'event':
            _session_refresh_favorite_events(self.request.session, self.request.user)
        elif request.POST['type'] == 'project':
            _session_refresh_favorite_projects(self.request.session, self.request.user)

        redirect_to = request.POST.get('next') or reverse('plainui:userprofile')
        url_is_safe = url_has_allowed_host_and_scheme(
            url=redirect_to,
            allowed_hosts=[self.request.get_host()],
            require_https=self.request.is_secure(),
        )
        if not url_is_safe:
            redirect_to = reverse('plainui:userprofile')
        return redirect(redirect_to)
